#!/usr/bin/env python

import os, argparse
import numpy as np
import denss


def store_parameters_as_string(sasrec):
    param_str = ("Parameter Values:\n"
    "Dmax  = {dmax:.5e}\n"
    "alpha = {alpha:.5e}\n"
    "Rg    = {rg:.5e} +- {rgerr:.5e}\n"
    "I(0)  = {I0:.5e} +- {I0err:.5e}\n"
    "Vp    = {Vp:.5e} +- {Vperr:.5e}\n"
    "MW_Vp = {mwVp:.5e} +- {mwVperr:.5e}\n"
    "MW_Vc = {mwVc:.5e} +- {mwVcerr:.5e}\n"
    "Lc    = {lc:.5e} +- {lcerr:.5e}\n"
    ).format(dmax=sasrec.D,alpha=sasrec.alpha,rg=sasrec.rg,rgerr=sasrec.rgerr,
        I0=sasrec.I0,I0err=sasrec.I0err,Vp=sasrec.Vp,Vperr=sasrec.Vperr,
        mwVp=sasrec.mwVp,mwVperr=sasrec.mwVperr,mwVc=sasrec.mwVc,mwVcerr=sasrec.mwVcerr,
        lc=sasrec.lc,lcerr=sasrec.lcerr)
    return param_str

def parse_arguments(parser):

    parser.add_argument("--version", action="version",version="%(prog)s v{version}".format(version=denss.__version__))
    parser.add_argument("-f", "--file", type=str, help="SAXS data file for input (either .dat, .fit, or .out)")
    parser.add_argument("-u", "--units", default="a", type=str, help="Angular units (\"a\" [1/angstrom] or \"nm\" [1/nanometer]; default=\"a\")")
    parser.add_argument("-m", "--mode", default="SLOW", type=str, help="Mode. F(AST) sets default options to run quickly for simple particle shapes. S(LOW) useful for more complex molecules. M(EMBRANE) mode allows for negative contrast. (default SLOW)")
    parser.add_argument("-d", "--dmax", default=None, type=float, help="Estimated maximum dimension")
    parser.add_argument("-n", "--nsamples", default=None, type=int, help="Number of samples, i.e. grid points, along a single dimension. (Sets voxel size, overridden by --voxel. Best optimization with n=power of 2. Default=64)")
    parser.add_argument("-ncs", "--ncs", default=0, type=int, help="Rotational symmetry")
    parser.add_argument("-ncs_steps","--ncs_steps", default=[3000,5000,7000,9000], nargs='+', help="Space separated list of steps for applying NCS averaging (default=3000 5000 7000 9000)")
    parser.add_argument("-ncs_axis", "--ncs_axis", default="1", type=str, help="Rotational symmetry axis (options: 1, 2, or 3 corresponding to (long,medium,short) principal axes)")
    parser.add_argument("-ncs_type", "--ncs_type", default="C", type=str, help="Symmetry type, either cyclical (default) or dihedral (i.e. C or D, dihedral (Dn) adds n 2-fold perpendicular axes)")
    parser.add_argument("-s", "--steps", default=None, help="Maximum number of steps (iterations)")
    parser.add_argument("-o", "--output", default=None, help="Output filename prefix")
    parser.add_argument("-v", "--voxel", default=None, type=float, help="Set desired voxel size, setting resolution of map")
    parser.add_argument("-os","--oversampling", default=3., type=float, help="Sampling ratio")
    parser.add_argument("--ne", default=None, type=float, help="Number of electrons in object (default 10000, set to negative number to ignore.)")
    parser.add_argument("--seed", default=None, help="Random seed to initialize the map")
    parser.add_argument("-rc","-rc_on", "--recenter_on", dest="recenter", action="store_true", help="Recenter electron density when updating support. (default)")
    parser.add_argument("-rc_off", "--recenter_off", dest="recenter", action="store_false", help="Do not recenter electron density when updating support.")
    parser.add_argument("-rc_steps", "--recenter_steps", default=None, type=int, nargs='+', help="List of steps to recenter electron density.")
    parser.add_argument("-rc_mode", "--recenter_mode", default="com", type=str, help="Recenter based on either center of mass (com, default) or maximum density value (max)")
    parser.add_argument("-p","-p_on","--positivity_on", dest="positivity", action="store_true", help="Enforce positivity restraint inside support. (default)")
    parser.add_argument("-p_off","--positivity_off", dest="positivity", action="store_false", help="Do not enforce positivity restraint inside support.")
    parser.add_argument("-p_steps", "--positivity_steps", default=None, type=int, nargs='+', help=argparse.SUPPRESS) #help="List of steps to enforce positivity.")
    parser.add_argument("-rho", "--rho_start", default=None, type=str, help="Starting electron density map filename (for use in denss_refine.py only)")
    parser.add_argument("-support", "--support", "--support_start", dest="support_start", default=None, type=str, help=argparse.SUPPRESS) #help="Starting electron density map filename of initial support (for use in denss_refine.py only)")
    parser.add_argument("--add_noise", default=None, type=float, help="Add noise to starting density map. Uniformly distributed random density is added to each voxel, by default from 0 to 1. The argument is a scale factor to multiply this by.")
    parser.add_argument("-e","-e_on","--extrapolate_on", dest="extrapolate", action="store_true", help=argparse.SUPPRESS) # help="Extrapolate data by Porod law to high resolution limit of voxels. (default)")
    parser.add_argument("-e_off","--extrapolate_off", dest="extrapolate", action="store_false", help=argparse.SUPPRESS) #help="Do not extrapolate data by Porod law to high resolution limit of voxels.")
    parser.add_argument("-sw","-sw_on","--shrinkwrap_on", dest="shrinkwrap", action="store_true", help="Turn shrinkwrap on (default)")
    parser.add_argument("-sw_off","--shrinkwrap_off", dest="shrinkwrap", action="store_false", help="Turn shrinkwrap off")
    parser.add_argument("-sw_om","-sw_old","-sw_om_on","--shrinkwrap_old_method_on", dest="shrinkwrap_old_method", action="store_true", help="Use the old method of shrinkwrap.")
    parser.add_argument("-sw_om_off","--shrinkwrap_old_method_off", dest="shrinkwrap_old_method", action="store_false", help="Use the new method of shrinkwrap (default).")
    parser.add_argument("-sw_start","--shrinkwrap_sigma_start_in_A", default=None, type=float, help="Starting sigma for Gaussian blurring, in angstroms")
    parser.add_argument("-sw_end","--shrinkwrap_sigma_end_in_A", default=None, type=float, help="Ending sigma for Gaussian blurring, in angstroms")
    parser.add_argument("-sw_start_vox","--shrinkwrap_sigma_start_in_vox", default=None, type=float, help="Starting sigma for Gaussian blurring, in voxels")
    parser.add_argument("-sw_end_vox","--shrinkwrap_sigma_end_in_vox", default=None, type=float, help="Ending sigma for Gaussian blurring, in voxels")
    parser.add_argument("-sw_decay","--shrinkwrap_sigma_decay", default=0.99, type=float, help="Rate of decay of sigma, fraction")
    parser.add_argument("-sw_threshold","--shrinkwrap_threshold_fraction", default=None, type=float, help="Minimum threshold defining support, in fraction of maximum density")
    parser.add_argument("-sw_iter","--shrinkwrap_iter", default=20, type=int, help="Number of iterations between updating support with shrinkwrap")
    parser.add_argument("-sw_minstep","--shrinkwrap_minstep", default=None, type=int, help="First step to begin shrinkwrap")
    parser.add_argument("-ec","-ec_on","--enforce_connectivity_on", dest="enforce_connectivity", action="store_true", help="Enforce connectivity of support, i.e. remove extra blobs (default)")
    parser.add_argument("-ec_off","--enforce_connectivity_off", dest="enforce_connectivity", action="store_false", help="Do not enforce connectivity of support")
    parser.add_argument("-ec_steps","--enforce_connectivity_steps", default=None, type=int, nargs='+', help="List of steps to enforce connectivity")
    parser.add_argument("-ec_max","--enforce_connectivity_max_features", default=1, type=int, help="Maximum number of features (i.e. disconnected blobs) allowed in support during enforce_connectivity step.")
    parser.add_argument("-cef", "--chi_end_fraction", default=0.001, type=float, help="Convergence criterion. Minimum threshold of chi2 std dev, as a fraction of the median chi2 of last 100 steps.")
    parser.add_argument("--write_xplor_format", default=False, action="store_true", help="Write out XPLOR map format (default only write MRC format).")
    parser.add_argument("--write_freq", default=100, type=int, help="How often to write out current density map (in steps, default 100).")
    parser.add_argument("--cutout_on", dest="cutout", action="store_true", help="When writing final map, cut out the particle to make smaller files.")
    parser.add_argument("--cutout_off", dest="cutout", action="store_false", help="When writing final map, do not cut out the particle to make smaller files (default).")
    parser.add_argument("--plot_on", dest="plot", action="store_true", help="Create simple plots of results (requires Matplotlib, default if module exists).")
    parser.add_argument("--plot_off", dest="plot", action="store_false", help="Do not create simple plots of results. (Default if Matplotlib does not exist)")
    parser.add_argument("-q", "--quiet", action="store_true", help="Do not display running statistics. (default False)")
    parser.add_argument("-gpu", "--gpu", dest="DENSS_GPU", action="store_true", help="Use GPU acceleration (requires CuPy). (default False)")
    parser.set_defaults(shrinkwrap=None)
    parser.set_defaults(shrinkwrap_old_method=None)
    parser.set_defaults(recenter=None)
    parser.set_defaults(positivity=None)
    parser.set_defaults(extrapolate=True)
    parser.set_defaults(enforce_connectivity=None)
    parser.set_defaults(cutout=False)
    parser.set_defaults(quiet = False)
    parser.set_defaults(DENSS_GPU = False)
    parser.set_defaults(plot=True)
    args = parser.parse_args()

    if args.plot:
        #if plotting is enabled, try to import matplotlib
        #if import fails, set plotting to false
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            args.plot = False

    if args.output is None:
        fname_nopath = os.path.basename(args.file)
        basename, ext = os.path.splitext(fname_nopath)
        if args.rho_start is not None:
            args.output = basename + "_refine"
        else:
            args.output = basename
    else:
        args.output = args.output

    #A bug appears to be present when disabling the porod extrapolation.
    #for now that option will be disabled until I come up with a fix
    if args.extrapolate is False:
        print ("There is currently a bug when disabling the Porod "
               "extrapolation (the -e_off option). \n For now, extrapolation "
               "has been re-enabled until a bug fix is released. ")
        args.extrapolate = True

    q, I, sigq, Ifit, file_dmax, isfit = denss.loadProfile(args.file, units=args.units)
    Iq = np.zeros((q.size,3))
    #for denss, I is actually the fit, since we want the smooth data
    #for reconstructions
    #also store the raw data for plotting
    Iq[:,0] = q
    Iq[:,1] = Ifit
    Iq[:,2] = sigq

    idx = np.where(I>0)
    qraw = np.copy(q)
    Iraw = np.copy(I)
    sigqraw = np.copy(sigq)
    idx = np.where(I>0)
    qraw = qraw[idx]
    Iraw = Iraw[idx]
    sigqraw = sigqraw[idx]

    if Iq.shape[0] < 3:
        print("Not enough data points (check that data has 3 columns: q, I, errors).")
        exit()

    Iq = denss.clean_up_data(Iq)
    is_raw_data = denss.check_if_raw_data(Iq)
    #now that we've cleaned up the data, reset the q, I, sigq arrays
    q = Iq[:,0]
    I = Iq[:,1]
    sigq = Iq[:,2]

    if args.dmax is not None and args.dmax >= 0:
        dmax = args.dmax
    elif file_dmax < 0:
        #if dmax from loadProfile() is -1, then dmax was not able
        #to be extracted from the file
        #in that case, estimate dmax directly from the data
        dmax, sasrec = denss.estimate_dmax(Iq)
    else:
        dmax = file_dmax
    D = dmax

    if is_raw_data:
        #in case a user gives raw experimental data, first, fit the data
        #using Sasrec and dmax
        sasrec = denss.Sasrec(Iq, D, alpha=0.0)
        ideal_chi2 = sasrec.calc_chi2()
        al = []
        chi2 = []
        #here, alphas are actually the exponents, since the range can
        #vary from 10^-10 upwards of 10^20. This should cover nearly all likely values
        alphas = np.arange(-10,20.)
        for alpha in alphas:
            #print("***** ALPHA ****** %.5e"%alpha)
            sasrec = denss.Sasrec(Iq, D, alpha=10.**alpha)
            r = sasrec.r
            pi = np.pi
            N = sasrec.N[:,None]
            In = sasrec.In[:,None]
            chi2value = sasrec.calc_chi2()
            al.append(alpha)
            chi2.append(chi2value)
        chi2 = np.array(chi2)

        #find optimal alpha value based on where chi2 begins to rise, to 10% above the ideal chi2 (where alpha=0)
        x = np.linspace(alphas[0],alphas[-1],1000)
        y = np.interp(x, alphas, chi2)
        chif = 2.0
        ali = np.argmin(y<=chif*ideal_chi2)
        opt_alpha = 10.0**(np.interp(chif*ideal_chi2,[y[ali+1],y[ali]],[x[ali+1],x[ali]])-1)
        alpha = opt_alpha

        sasrec = denss.Sasrec(Iq, D, alpha=alpha)
        #sasrec = denss.Sasrec(Iq, dmax, qc=None, extrapolate=False)
        #now, set the Iq values to be the new fitted q values
        q = sasrec.qc
        I = sasrec.Ic
        sigq = sasrec.Icerr

        #save fit, just like from denss_fit_data.py
        param_str = store_parameters_as_string(sasrec)
        #add column headers to param_str for output
        param_str += 'q, I, error, fit'
        #quick, interpolate the raw data, sasrec.I, to the new qc values, but be sure to 
        #put zeros in for the q values not measured behind the beamstop
        Iinterp = np.interp(sasrec.qc, sasrec.q, sasrec.I, left=0.0, right=0.0)
        np.savetxt(args.output+'_fitdata.fit', np.vstack((sasrec.qc, Iinterp, sasrec.Icerr, sasrec.Ic)).T,delimiter=' ',fmt='%.5e',header=param_str)

    #denss cannot deal with negative intensity values which sometimes happens with real data
    #if the fit goes negative, denss will fail
    #so first, remove all negative intensity values from the fit
    idx_pos = np.where(I>0)
    q = q[idx_pos]
    I = I[idx_pos]
    sigq = sigq[idx_pos]

    #allow ncs_steps to be either list of ints or string of list of ints
    if isinstance(args.ncs_steps, list):
        if len(args.ncs_steps) == 1:
            args.ncs_steps = np.fromstring(args.ncs_steps[0],sep=' ',dtype=int)
        else:
            args.ncs_steps = [int(x) for x in args.ncs_steps]

    if args.ncs_type[0].upper() == "D":
        args.ncs_type = "dihedral"
    else:
        args.ncs_type = "cyclical"

    if args.ncs_axis[0].upper() == "L" or args.ncs_axis[0] == "1":
        #if "1" or "long" or "LONG" or "L" or similar is given
        #assume the long axis, i.e. axis = 1
        args.ncs_axis = 1
    elif args.ncs_axis[0].upper() == "M" or args.ncs_axis[0] == "2":
        #if "2" or "middle" or "MIDDLE" or "M" or similar is given
        #assume the middle axis, i.e. axis = 2
        args.ncs_axis = 2
    elif args.ncs_axis[0].upper() == "S" or args.ncs_axis[0] == "3":
        #if "3" or "short" or "SHORT" or "S" or similar is given
        #assume the short axis, i.e. axis = 3
        args.ncs_axis = 3
    else:
        #default to long axis
        args.ncs_axis = 1

    #old default sw_start was 3.0
    #however, in cases where the voxel size is smaller than default,
    #e.g. increasing nsamples, shrinkwrap works differently since sigma
    #is then smaller (in angstroms). We will instead base sigma on
    #physical dimension in angstroms, rather than number of voxels.
    #We will do this by basing sigma on particle size (dmax), rather than 
    #resolution of the grid. 
    #since defaults work well, we'll start there. The parameters
    #defining voxel size are the grid size / nsamples (dmax*oversampling/nsamples)
    #so define sw_start as (voxsize*3) = (3*D/64) * 3, since those are the defaults that work well
    #later we need to convert this physical size into a voxel size (since 
    #ndimage.gaussian_filter works based on number of pixels)

    #some defaults:
    shrinkwrap_sigma_start_in_A = (3.0 * dmax / 64.0) * 3.0
    shrinkwrap_sigma_end_in_A = (3.0 * dmax / 64.0) * 1.5
    shrinkwrap_threshold_fraction = 0.2
    positivity = True

    if args.shrinkwrap_old_method:
        #for FAST or SLOW modes, set some default values for a few options
        if args.mode[0].upper() == "F":
            args.mode = "FAST"
            nsamples = 32
            shrinkwrap_minstep = 0
            ec_steps_to_add = np.array([1000])
            enforce_connectivity_steps = shrinkwrap_minstep + ec_steps_to_add
            recenter_steps = list(range(501,2502,500))
        elif args.mode[0].upper() == "S":
            args.mode = "SLOW"
            nsamples = 64
            shrinkwrap_minstep = 0
            ec_steps_to_add = np.array([1000])
            enforce_connectivity_steps = shrinkwrap_minstep + ec_steps_to_add
            recenter_steps = list(range(501,8002,500))
        elif args.mode[0].upper() == "M":
            args.mode = "MEMBRANE"
            nsamples = 64
            positivity = False
            shrinkwrap_minstep = 0
            ec_steps_to_add = np.array([300])
            enforce_connectivity_steps = shrinkwrap_minstep + ec_steps_to_add
            recenter_steps = list(range(501,8002,500))
        else:
            args.mode = "None"
    else:
        #for FAST or SLOW modes, set some default values for a few options
        if args.mode[0].upper() == "F":
            args.mode = "FAST"
            nsamples = 32
            shrinkwrap_minstep = 0
            ec_steps_to_add = np.array([1000])
            enforce_connectivity_steps = shrinkwrap_minstep + ec_steps_to_add
            recenter_steps = list(range(501,2502,500))
        elif args.mode[0].upper() == "S":
            args.mode = "SLOW"
            nsamples = 64
            shrinkwrap_minstep = 0
            ec_steps_to_add = np.array([1000])
            enforce_connectivity_steps = shrinkwrap_minstep + ec_steps_to_add
            recenter_steps = list(range(501,8002,500))
        elif args.mode[0].upper() == "M":
            args.mode = "MEMBRANE"
            nsamples = 64
            positivity = False
            shrinkwrap_minstep = 0
            shrinkwrap_sigma_start_in_A *= 2.0
            shrinkwrap_sigma_end_in_A *= 2.0
            ec_steps_to_add = np.array([300, 500, 1000])
            enforce_connectivity_steps = shrinkwrap_minstep + ec_steps_to_add
            recenter_steps = list(range(501,8002,500))
        else:
            args.mode = "None"

    #allow user to explicitly modify those values by resetting them here to the user defined values
    if args.nsamples is not None:
        nsamples = args.nsamples

    if args.voxel is None and nsamples is None:
        voxel = 5.
    elif args.voxel is None and nsamples is not None:
        voxel = dmax * args.oversampling / nsamples
    else:
        voxel = args.voxel

    if args.positivity is not None:
        positivity = args.positivity

    if args.shrinkwrap_minstep is not None:
        shrinkwrap_minstep = args.shrinkwrap_minstep
        #adjust ec_steps if minstep is given by user by default
        #ec_steps will be overwritten later if user defined
        enforce_connectivity_steps = shrinkwrap_minstep + ec_steps_to_add

    if args.shrinkwrap_threshold_fraction is not None:
        shrinkwrap_threshold_fraction = args.shrinkwrap_threshold_fraction

    #allow user to input sigma as angstroms or voxels
    if args.shrinkwrap_sigma_start_in_A is not None:
        shrinkwrap_sigma_start_in_A = args.shrinkwrap_sigma_start_in_A
    elif args.shrinkwrap_sigma_start_in_vox is not None:
        #if voxel option is set, for now convert to angstroms, 
        #then later convert everything back to voxels
        shrinkwrap_sigma_start_in_A = args.shrinkwrap_sigma_start_in_vox * voxel

    if args.shrinkwrap_sigma_end_in_A is not None:
        shrinkwrap_sigma_end_in_A = args.shrinkwrap_sigma_end_in_A
    elif args.shrinkwrap_sigma_end_in_vox is not None:
        shrinkwrap_sigma_end_in_A = args.shrinkwrap_sigma_end_in_vox * voxel

    #as mentioned above, now that we have the voxel size, we need to convert
    #the shrinkwrap sigma values to voxels, rather than physical distance
    shrinkwrap_sigma_start_in_vox = shrinkwrap_sigma_start_in_A / voxel
    shrinkwrap_sigma_end_in_vox = shrinkwrap_sigma_end_in_A / voxel

    if args.enforce_connectivity_steps is not None:
        enforce_connectivity_steps = args.enforce_connectivity_steps
    if not isinstance(enforce_connectivity_steps, np.ndarray):
        enforce_connectivity_steps = np.asarray(enforce_connectivity_steps)

    if args.recenter_steps is not None:
        recenter_steps = args.recenter_steps
    if not isinstance(recenter_steps, list):
        recenter_steps = [ recenter_steps ]

    if args.positivity_steps is not None:
        print("The positivity_steps option is currently not allowed.")
        # if not isinstance(args.positivity_steps, list):
        #     args.positivity_steps = [ positivity_steps ]

    if args.steps is not None:
        steps = args.steps

    oversampling = args.oversampling
    side = dmax * oversampling

    #allow user to give initial density map for denss_refine.py
    if args.rho_start is not None:
        rho_start, rho_side = denss.read_mrc(args.rho_start)
        rho_nsamples = rho_start.shape[0]
        rho_voxel = rho_side/rho_nsamples

        if (not np.isclose(rho_side, side) or
            not np.isclose(rho_voxel, voxel) or
            not np.isclose(rho_nsamples, nsamples)):
            print("rho_start density dimensions do not match given options.")
            print("Oversampling and voxel size adjusted to match rho_start dimensions.")
            side = rho_side
            voxel = rho_voxel
            oversampling = side/dmax
            nsamples = rho_nsamples
        #set args.rho_start to the actual density map, rather than the filename
        args.rho_start = rho_start
        if args.recenter is None:
            args.recenter = False
        if args.shrinkwrap_old_method is None:
            shrinkwrap_old_method = True
        else:
            shrinkwrap_old_method = args.shrinkwrap_old_method

    #allow user to give initial support, check for consistency with given grid parameters
    if args.support_start is not None:
        support_start, support_side = denss.read_mrc(args.support_start)
        support_nsamples = support_start.shape[0]
        support_voxel = support_side/support_nsamples

        if (not np.isclose(support_side, side) or
            not np.isclose(support_voxel, voxel) or
            not np.isclose(support_nsamples, nsamples)):
            print("Support dimensions do not match given options.")
            print("Ignoring support.")
            print("side (support, given): ", support_side, side)
            print("voxel (support, given): ", support_voxel, voxel)
            print("n (support, given): ", support_nsamples, nsamples)
            args.support_start = None
        else:
            args.support_start = support_start.astype(bool)

    if args.shrinkwrap is None:
        if args.support_start is not None:
            #assume if a user gives a support, that they do not want to
            #run shrinkwrap. However, allow the user to enable shrinkwrap explicitly
            #by setting the shrinkwrap option on
            args.shrinkwrap = False
            if args.enforce_connectivity is None:
                args.enforce_connectivity = False
            if args.recenter is None:
                args.recenter = False
        else:
            args.shrinkwrap = True

    if args.enforce_connectivity is None:
        args.enforce_connectivity = True
    if args.recenter is None:
        args.recenter = True
    if args.shrinkwrap_old_method is None:
        args.shrinkwrap_old_method = False

    if args.ne is not None:
        if args.ne <= 0.0:
            #if args.ne is negative, then assume that the user
            #does not want to set the number of electrons, and just
            #set it equal to the square root of the forward scattering I(0)
            args.ne = I[0]**0.5
    elif args.rho_start is not None:
        #if args.ne is not given, and args.rho_start is given,
        #then set ne to be the sum of the given density map
        dV = voxel**3
        args.ne = args.rho_start.sum() * dV
    else:
        #default to 10,000
        args.ne = 10000.0

    #now recollect all the edited options back into args
    args.nsamples = nsamples
    args.positivity = positivity
    args.shrinkwrap_minstep = shrinkwrap_minstep
    args.shrinkwrap_threshold_fraction = shrinkwrap_threshold_fraction
    args.shrinkwrap_sigma_start = shrinkwrap_sigma_start_in_vox
    args.shrinkwrap_sigma_end = shrinkwrap_sigma_end_in_vox
    args.enforce_connectivity_steps = enforce_connectivity_steps
    args.recenter_steps = recenter_steps
    args.oversampling = oversampling
    args.dmax = dmax
    args.voxel = voxel
    args.q = q
    args.I = I
    args.sigq = sigq
    args.qraw = qraw
    args.Iraw = Iraw
    args.sigqraw = sigqraw

    return args



# coding: utf-8
"""

    BasePercentileOffset.py

    scattering baseline solver assuming Gaussian Curves

    Copyright (c) 2017-2020, Masatsuyo Takahashi, KEK-PF

"""
import os
import numpy            as np
from scipy.stats        import norm
from bisect             import bisect_right

TABLE_SIZE          = 100
N_SCALE             = 0.01
MAX_NOISINESS       = 1.0
START_SIZE_SIGMA    = 3
END_SIZE_SIGMA      = 13
NUM_SIZE_SIGMAS     = 21
DEFAULT_OFFSET      = 10

def compute_noise_and_size_sigma_dependency( num_iteretions=1000 ):
    N = 300
    np.random.seed(1234)

    ym  = norm.pdf( 0, 0, 1 )
    scale = 1/ym

    dat_list = []
    std_list = []

    for n in range(TABLE_SIZE):
        nn = N_SCALE * (n+1)

        p_list = [ nn ]
        e_list = [ nn ]

        for size_sigma in np.linspace( START_SIZE_SIGMA, END_SIZE_SIGMA, NUM_SIZE_SIGMAS ):
            x   = np.linspace( -size_sigma, size_sigma, N )
            y0  = norm.pdf( x, 0, 1 )
            y_  = y0 * scale

            ys_array = []
            zi_array = []

            for i in range(num_iteretions):
                noise = np.random.normal(0, nn, N)
                y   = y_ + noise
                ys  = sorted( y )
                zi  = bisect_right( ys, 0 )
                if False:
                    import molass_legacy.KekLib.DebugPlot as plt
                    if i==0 and n%10==0 and abs(size_sigma - 9) < 0.1:
                        plt.push()
                        fig, ax = plt.subplots()
                        axt = ax.twinx()
                        ax.set_title("nn=%.3g, %.3g sigma width" % (nn, size_sigma))
                        ax.plot(y_)
                        ax.plot(y)
                        axt.plot(ys, ':')
                        axt.plot(zi, ys[zi], 'o', color='red')
                        fig.tight_layout()
                        plt.show()
                        plt.pop()
                zi_array.append( zi )
                ys_array.append( ys[zi] )

            p = np.average( zi_array ) * 100 / N
            e = np.std( zi_array )
            p_list.append( p )
            e_list.append( e )

        dat_list.append( p_list )
        std_list.append( e_list )

    return np.array(dat_list), np.array(std_list)

bpo_table = None
std_table = None

def load_bpo_table():
    global bpo_table
    if bpo_table is not None: return

    folder = os.path.dirname( __file__ )
    bpo_dat_path = folder + '/base_percentile_offset.dat'
    # bpo_std_path = folder + '/base_percentile_offset_std.dat'
    if not os.path.exists( bpo_dat_path ):
        dat, std = compute_noise_and_size_sigma_dependency( num_iteretions=1000 )
        np.savetxt( bpo_dat_path, dat )
        # np.savetxt( bpo_std_path, std )
        bpo_table = dat
        # std_table = std
    else:
        # import time
        # start_ = time.time()
        bpo_table = np.loadtxt( bpo_dat_path )
        # end_ = time.time()
        # std_table = np.loadtxt( bpo_std_path )
        # print( 'took', end_ - start_, 'to load bpo_table.' )

    # print('bpo_table.shape=', bpo_table.shape)

load_bpo_table()

def base_percentile_offset( noisiness, size_sigma=5 ):
    # See Also ElutionBaseCurve.py
    # return DEFAULT_OFFSET

    if not np.isfinite( noisiness ):
        noisiness   = MAX_NOISINESS

    nn = max( 0, min( TABLE_SIZE - 1, int( round( noisiness / N_SCALE ) ) ) )
    j = max( 0, min( NUM_SIZE_SIGMAS - 1, int( round( ( size_sigma - START_SIZE_SIGMA )*2 ) ) ) )
    # print('nn, j=', (nn, j))
    bpo =  bpo_table[ nn, j+1 ]
    # print( 'base_percentile_offset: noisiness=', noisiness, 'size_sigma=', size_sigma, 'bpo=', bpo  )
    return bpo



"""
    Optimizer.ComplementaryView.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from bisect import bisect_right
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy._MOLASS.SerialSettings import get_setting
from DataUtils import get_in_folder
from molass_legacy.KekLib.BasicUtils import ordinal_str
from .LrfExporter import LrfExporter
from molass_legacy.SerialAnalyzer.AnalyzerUtil import compute_conc_factor_util
from molass_legacy.Saxs.SaxsCurveUtils import percentile_normalize
from molass_legacy._MOLASS.Version import is_developing_version

TOGGLE_BUTTON_TEXTS = ["Show", "Hide"]

class ComplementaryView(Dialog):
    def __init__(self, parent, optimizer, curr_index, params, work_folder=None, sd=None):
        self.devel = is_developing_version()
        self.parent = parent
        self.optimizer = optimizer
        self.xr_curve = optimizer.xr_curve
        self.uv_curve = optimizer.uv_curve
        self.xrD = optimizer.xrD
        self.uvD = optimizer.uvD
        self.xrD_ = optimizer.xrD_
        self.uvD_ = optimizer.uvD_
        self.qv = optimizer.qvector
        self.wv = optimizer.wvector
        self.curr_index = curr_index
        self.params = params
        self.separate_params = optimizer.split_params_simple(params)
        self.lrf_info = optimizer.objective_func(params, return_lrf_info=True)
        self.work_folder = work_folder
        self.sd = sd
        self.show_baseline = 0
        self.separate_eoii = optimizer.separate_eoii
        self.popup_menu = None
        Dialog.__init__(self, parent, "Complementary View", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        # fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(21, 11))
        fig = plt.figure(figsize=(18, 9))
        gs = GridSpec(2, 10)
        ax0 = fig.add_subplot(gs[0,0])
        ax1 = fig.add_subplot(gs[1,0])
        self.label_axes = (ax0, ax1)
        axes_list = []
        for i in range(2):
            row_list = []
            for j in range(3):
                jb = j*3 + 1
                row_list.append(fig.add_subplot(gs[i,jb:jb+3]))
            axes_list.append(row_list)
        axes = np.array(axes_list)
        self.fig = fig
        self.axes = axes
        self.axt = axes[1,0].twinx()
        self.draw()
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.draw()
        if self.devel:
            fig.canvas.mpl_connect('button_press_event', self.on_click)

    def buttonbox(self):
        lower_frame = Tk.Frame(self)
        lower_frame.pack(fill=Tk.BOTH, expand=1)

        width = int(self.mpl_canvas_widget.cget('width'))
        padx = width*0.03

        tframe = Tk.Frame(lower_frame)
        tframe.pack(side=Tk.LEFT, padx=padx)
        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        space = Tk.Frame(lower_frame, width=padx)
        space.pack(side=Tk.RIGHT)

        box = Tk.Frame(lower_frame)
        box.pack(side=Tk.RIGHT)

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=padx, pady=5)

        text = self.get_toggle_button_text()
        w = Tk.Button(box, text=text, width=12, command=self.toggle_baseline)
        w.pack(side=Tk.LEFT, padx=padx, pady=5)
        self.toggle_btn = w

        w = Tk.Button(box, text="Guinier-Kratky View", width=16, command=self.show_guinier_kratky_view)
        w.pack(side=Tk.LEFT, padx=padx, pady=5)

        w = Tk.Button(box, text="Compare to PDB (CRYSOL)", width=22, command=self.show_compare_to_pdb)
        w.pack(side=Tk.LEFT, padx=padx, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def get_toggle_button_text(self):
        return "%s Baseline" % TOGGLE_BUTTON_TEXTS[self.show_baseline]

    def cancel(self):
        plt.close(self.fig)
        Dialog.cancel(self)

    def draw(self):
        self.draw_axis_labels()
        self.draw_complements()

    def draw_axis_labels(self):
        for ax in self.label_axes:
            ax.cla()
            ax.set_axis_off()
        ax0, ax1 = self.label_axes
        ax0.text(0.5, 0.5, "UV", ha="center", fontsize=20)
        ax1.text(0.5, 0.5, "Xray", ha="center", fontsize=20)

    def draw_complements(self):
        fig = self.fig
        axes = self.axes
        for ax_row in axes:
            for ax in ax_row:
                ax.cla()
        self.axt.cla()
        self.axt.grid(False)

        in_folder = get_in_folder()
        fig.suptitle("Complement View at %s local minimum on %s" % (ordinal_str(self.curr_index), in_folder), fontsize=20)
        ax0, ax1, ax2 = axes[0,:]
        ax0.set_title("Used Elution Components", fontsize=16)
        ax1.set_title("Linear Scale", fontsize=16)
        ax2.set_title("Log Scale", fontsize=16)

        ax00 = axes[0,0]
        ax10 = axes[1,0]
        self.optimizer.objective_func(self.params, plot=True, axis_info=(self.fig, [ax00, ax10, None, self.axt]))
        ax00.set_xlabel(r"$\lambda$")
        ax00.set_ylabel("Absorbance")
        ax10.set_xlabel("Q")
        ax10.set_ylabel("Intensity")

        lrf_info = self.lrf_info
        Pxr, Cxr, Puv, Cuv, mapped_Uv = lrf_info.matrices
        self.draw_complement_impl(axes[0,1:], self.wv, Puv, Cuv,
                color="blue", value_label="Absorbance", vector_label=r"$\lambda$")

        conc_factor = compute_conc_factor_util()
        Cxr_ = Cxr/conc_factor
        self.draw_complement_impl(axes[1,1:], self.qv, Pxr, Cxr_,
                color="orange", value_label="Intensity (max normalized)", vector_label="Q", xray=True)

        self.adjust_xlim_mapping(axes[0,0], axes[1,0])

        fig.tight_layout()
        fig.subplots_adjust(top=0.9)

    def draw_complement_impl(self, axes, vector, P, C, color=None, value_label="", vector_label="", xray=False):
        if not xray:
            if False:
                import UV.Foldedness
                from importlib import reload
                reload(UV.Foldedness)
            from molass_legacy.UV.Foldedness import Foldedness
            foldedness = Foldedness(self.wv)

        num_others = 2 if xray and self.separate_eoii else 1
        num_components = len(C) - num_others
        draw_baseline = self.show_baseline == 1

        PT = P.T[0:num_components+1]
        if xray:
            PT = [percentile_normalize(v) for v in PT]

        for k, ax in enumerate(axes):
            if k == 1:
                if xray:
                    ax.set_yscale('log')
                else:
                    ax.set_axis_off()
                    ax.text(0.5, 0.5, "Suppressed", fontsize=20, alpha=0.3, ha="center", va="center")
                    continue

            for j, v in enumerate(PT):
                if j < num_components:
                    if xray:
                        fnstr = ""
                    else:
                        fnstr = ", FnR=%.1f" % foldedness.compute(v)
                    ax.plot(vector, v, label="component-%d%s" % (j+1, fnstr), alpha=0.5)
                else:
                    if draw_baseline:
                        ax.plot(vector, v, color='red', label="baseline", alpha=0.5)

            if k == 0 and not xray:
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                for w, color in zip(foldedness.get_wavelengths(), foldedness.get_plotcolors()):
                    ax.plot([w, w], [ymin, ymax], ":", label="wavelength=%g" % w, color=color)

            if xray and self.separate_eoii:
                v = P.T[-1]
                ax.plot(vector, v, '-', color="pink", label="effect of ii", alpha=0.5)

            ax.set_xlabel(vector_label)
            ax.set_ylabel(value_label)
            ax.legend()

    def adjust_xlim_mapping(self, ax0, ax1):
        a, b = self.separate_params[3]
        xmin, xmax = [a*x + b for x in ax1.get_xlim()]
        ax0.set_xlim(xmin, xmax)

    def toggle_baseline(self):
        self.show_baseline = 1 - self.show_baseline
        self.draw()
        self.mpl_canvas.draw()
        text = self.get_toggle_button_text()
        self.toggle_btn.config(text=text)
        self.update()

    def show_guinier_kratky_view(self):
        try:
            self.show_guinier_kratky_view_impl()
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "show_guinier_kratky_view: ")

    def show_guinier_kratky_view_impl(self, debug=False):
        if debug:
            from importlib import reload
            import Kratky.GuinierKratkyView
            reload(Kratky.GuinierKratkyView)
        from Kratky.GuinierKratkyView import GuinierKratkyView

        dialog = GuinierKratkyView(self.parent, self.optimizer, self.params, self.lrf_info)
        dialog.show()

    def show_compare_to_pdb(self, devel=True):
        if devel:
            from importlib import reload
            import Theory.PdbCrysolRoute
            reload(Theory.PdbCrysolRoute)
        from Theory.PdbCrysolRoute import compare_to_pdb_impl
        compare_to_pdb_impl(self)

    def on_click(self, event):
        if event.xdata is None:
            return

        if event.button == 3:
            from molass_legacy.KekLib.TkUtils import split_geometry
            self.create_popup_menu(event)
            rootx = self.winfo_rootx()
            rooty = self.winfo_rooty()
            w, h, x, y = split_geometry(self.mpl_canvas_widget.winfo_geometry())
            self.popup_menu.post(rootx + int(event.x), rooty + h - int(event.y))
            return

    def create_popup_menu(self, event):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu(self, tearoff=0)
            self.popup_menu.add_command(label='Show Simple 3D View', command=self.show_simple_3d_view)

    def show_simple_3d_view(self, debug=True):
        if debug:
            import Tools.SimpleThreedView
            from importlib import reload
            reload(Tools.SimpleThreedView)
        from molass_legacy.Tools.SimpleThreedView import show_simple_3d_view
        show_simple_3d_view(self.sd)



# coding: utf-8
"""
    EFA.RotationDemo.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
# import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Polygon
import matplotlib.animation as animation
from CorrectedData import CorrectedXray
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.OurMatplotlib import get_color
from MatrixData import simple_plot_3d
from ModeledData import ModeledData
from Tutorials.MatrixFactorization import draw_PQ, generate_random_points
from Tutorials.SvdPrecision import plot_text

FONTSIZE = 160
A_SIZE = 600

def cpc_demo(symmetric=False):
    q = np.linspace(0.01, 0.6, A_SIZE)
    rg_list = [50, 24]
    d_list = [1, 3]

    if symmetric:
        E_SIZE = 300
        md = ModeledData(q, E_SIZE, rg_list=rg_list, d_list=d_list)
        f1, t1 = 0, 200
        f2, t2 = 100, E_SIZE
    else:
        E_SIZE = 500
        md = ModeledData(q, E_SIZE, rg_list=rg_list, d_list=d_list,
                            h_list=[1, 0.5], mu_list=[130, 220], sigma_list=[20, 30], tau_list=[40, 20])
        f1, t1 = 0, 350
        f2, t2 = 120, E_SIZE

    R = np.zeros(md.C.shape)
    R[0,f1:t1] = 1
    R[1,f2:t2] = 1

    def normalize_C(C):
        C = R*C
        for k in range(C.shape[0]):
            C[k,:] /= np.sum(C[k,:])
        return C

    def one_iteration(C):
        P = M_ @ np.linalg.pinv(normalize_C(C))
        C_ = np.linalg.pinv(P) @ M_
        return C_

    fig = plt.figure(figsize=(22,11))
    fig.suptitle("Early (1-4) C-P-C Iterations", fontsize=40)
    gs = GridSpec(3,6)
    axes = []
    for i in range(3):
        ax_row = []
        for j in range(6):
            ax = fig.add_subplot(gs[i,j], projection='3d' if i+j == 0 else None)
            ax_row.append(ax)
        axes.append(ax_row)

    rank = len(rg_list)
    axes = np.array(axes)

    ax00 = axes[0,0]
    M = md.get_data()
    U, s, VT = np.linalg.svd(M)
    M_ = U[:,0:rank] @ np.diag( s[0:rank] ) @ VT[0:rank,:]
    simple_plot_3d(ax00, M_, x=q)

    ax10 = axes[1,0]
    for p in md.P.T:
        ax10.plot(q, p)

    ax20 = axes[2,0]
    for c in md.C:
        ax20.plot(c)

    poly1 = Polygon([(f1,0), (f1,1), (t1,1), (t1,0)], color='cyan', alpha=0.2)
    ax20.add_patch(poly1)
    poly2 = Polygon([(f2,0), (f2,1), (t2,1), (t2,0)], color='pink', alpha=0.2)
    ax20.add_patch(poly2)

    def early_iterations(i, C):
        ax0 = axes[i,1]
        for c in C:
            ax0.plot(c)

        for j in range(2,6):
            ax = axes[i,j]
            C = one_iteration(C)
            for c in C:
                ax.plot(c)
            plot_text(ax, str(j-1), fontsize=FONTSIZE)

    early_iterations(0, VT[0:rank,:])
    plot_text(axes[0,1], "$V^T$", fontsize=FONTSIZE)

    w = np.linspace(0, 1, E_SIZE)
    W = np.vstack([w, 1-w])
    early_iterations(1, W)

    r = np.random.uniform(0, 1, E_SIZE)
    R = np.vstack([r, 1-r])
    early_iterations(2, R)

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)
    plt.show()



"""
    SimTools.HardSphere.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from Theory.SolidSphere import phi
from molass_legacy.Models.ElutionCurveModels import egha

def formfactor_factor(qv, R):
    return phi(qv, R)**2

def structure_factor(qv, K, R):
    return 1 - K*phi(qv, 2*R)

def get_model_data(qv, jv, rg, h, mu, sigma, tau, a=0, K=0, error=0, return_error_matrix=False):
    R = np.sqrt(5/3)*rg

    fy = formfactor_factor(qv, R)
    sy = structure_factor(qv, K, R)

    if error > 0:
        fy += np.random.normal(0, np.max(fy)*error, len(qv))
        sy += np.random.normal(0, np.max(np.abs(sy))*error, len(qv))

    aq = fy
    bq = fy*(sy - 1)
    P = np.array([aq, bq]).T

    cy = egha(jv, h, mu, sigma, tau, a)

    if error > 0:
        cy += np.random.normal(0, np.max(cy)*error, len(jv))

    C = np.array([cy, cy**2])

    M = P @ C

    if return_error_matrix:
        from SvdDenoise import get_denoised_error
        M_, P_, C_ = get_model_data(qv, jv, rg, h, mu, sigma, tau, a=a, K=K, error=0)
        E = M - M_
        E_ = get_denoised_error(M_, M, E)   # better method wanted
        return M, E_, P, C
    else:
        return M, P, C



# coding: utf-8
"""
    Optimizer.BoundsAnalysis.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
import glob
import molass_legacy.KekLib.DebugPlot as plt

def demo(parent, analysis_folder, trimming=True, correction=True):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Optimizer.TheUtils import get_work_info
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
    from DataUtils import get_in_folder
    from molass_legacy.Optimizer.StateSequence import read_callback_txt_impl

    set_setting("analysis_folder", analysis_folder)
    work_folder = analysis_folder + "/optimized/jobs/000"
    in_folder = get_work_info(work_folder).in_folder
    in_folder = in_folder.replace("E:", "D:")
    set_setting("in_folder", in_folder)
    if False:
        sp = StandardProcedure()
        sd = sp.load_old_way(in_folder)
        pre_recog = PreliminaryRecognition(sd)
        if trimming:
            sd_ = sd._get_analysis_copy_impl(pre_recog)
        else:
            sd_ = sd.get_copy()

        if correction:
            v2_copy = get_corrected_sd_impl(sd_, sd, pre_recog)
        else:
            v2_copy = sd_

        D, _, wv, ecurve = v2_copy.get_uv_data_separate_ly()

    for k, file in enumerate(glob.glob(analysis_folder + "/optimized/jobs/*/callback.txt")):
        print([k], file)
        fv_list, x_list = read_callback_txt_impl(file)

    with plt.Dp():
        in_folder = get_in_folder(in_folder)
        fig, ax = plt.subplots()
        ax.set_title("Bounds Analysis for %s" % in_folder, fontsize=20)
        # simple_plot(ax, ecurve, legend=False)
        ax.plot(0, 0, 'o')
        fig.tight_layout()
        plt.show()



"""
    Optimizer.ScaleAdjustInspect.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from .RatioInterpretIllust import MODEL_NAME_DICT
from .FvScoreConverter import convert_score

USE_BH = True

def scale_adjust_inspect(caller):
    print("scale_adjust_inspect")
    js_canvas = caller.canvas

    optimizer = js_canvas.fullopt
    func_code = optimizer.get_name()
    model_name = MODEL_NAME_DICT[func_code]
    assert model_name == "SDM"
    params = js_canvas.get_current_params()
    job_name = js_canvas.dialog.get_job_info()[0]
    in_folder = get_in_folder()

    scale_indeces_list = [[0,1,2], [8,9,10]]  # 0,1,2: xr scales, 8,9,10: uv scales
    
    if USE_BH:
        temp_params = optimizer.to_norm_params(params)
    else:
        temp_params = params.copy()

    for scale_indeces in scale_indeces_list:
        print("optimizing with scale_indeces", scale_indeces)
        def scale_objective(p):
            temp_params[scale_indeces] = p
            if USE_BH:
                return optimizer.objective_func_wrapper(temp_params)
            else:
                return optimizer.objective_func(temp_params)

        if USE_BH:
            minimizer_kwargs = dict(method='Nelder-Mead')
            res = basinhopping(scale_objective, temp_params[scale_indeces], niter=20, minimizer_kwargs=minimizer_kwargs)
        else:
            res = minimize(scale_objective, temp_params[scale_indeces], method="Nelder-Mead")
        temp_params[scale_indeces] = res.x

    if USE_BH:
        temp_params = optimizer.to_real_params(temp_params)

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(18, 10))
        ax1 = axes[0,0]
        ax2 = axes[0,1]
        ax2t = ax2.twinx()
        ax2t.grid(False)
        ax3 = axes[0,2]
        ax4 = axes[1,0]
        ax5 = axes[1,1]
        ax5t = ax5.twinx()
        ax5t.grid(False)
        ax6 = axes[1,2]
        fig.suptitle("Scale Adjust Inspection at %s with=%s of %s" % (job_name, model_name, in_folder), fontsize=20)
        fv1 = optimizer.objective_func(params, plot=True, axis_info=(fig, (ax1, ax2, ax3, ax2t)))
        fv2 = optimizer.objective_func(temp_params, plot=True, axis_info=(fig, (ax4, ax5, ax6, ax5t)))
        ax1.set_title("Current Result UV", fontsize=16)
        ax2.set_title("Current Result XR", fontsize=16)
        ax3.set_title("Current Result SV=%.3g" % convert_score(fv1), fontsize=16)
        ax4.set_title("Scale Adjusted UV", fontsize=16)
        ax5.set_title("Scale Adjusted XR", fontsize=16)
        ax6.set_title("Scale Adjusted SV=%.3g" % convert_score(fv2), fontsize=16)
        fig.tight_layout()
        plt.show()


"""
For tests
"""
import time
from model_gm import ModelGM, sample_gm
from em import EM
from mm import MM
from dmm import DMM
from discrete_rv import wass



if __name__ == '__main__':
    k = 2
    mm = MM(k, sigma=1)
    em = EM(k, sigma=1)
    dmm = DMM(k, sigma=None)

    model = ModelGM(w=[0.5, 0.5], x=[-.5, .5], std=1)
    sample = sample_gm(model, 10000)

    # esti_em = em.estimate(sample)
    # print(wass(esti_em.mean_rv(), model.mean_rv()))
    print(dmm.estimate(sample))
    print(dmm.estimate_online(sample))
    # print(wass(esti_dmm, model.mean_rv()))
    # print(mm.estimate(sample))

    # dmm = DMM(k=2)
    # print(dmm.estimate(sample))



# coding: utf-8
"""
    DelayAdjuster.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from lmfit import Parameters, minimize
import matplotlib.pyplot as plt
import matplotlib.patches as mpl_patches
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, is_empty_val
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable, BlinkingFrame
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy.KekLib.TkUtils import is_low_resolution
from molass_legacy._MOLASS.SerialSettings import get_setting, get_xray_picking
from DelayOptimizer import DelayOptimizer
from molass_legacy._MOLASS.SerialSettings import set_setting
from SvdDenoise import get_denoised_data

MODEL_ID_LIST = ["One State", "Two-State Unfolding", "Three-State Unfolding"]

class DelayAdjuster(Dialog):
    def __init__(self, parent, xdata, in_folder):
        self.logger = logging.getLogger(__name__)
        self.busy = False
        self.applied = False
        self.xdata = xdata
        self.mapper = None
        self.in_folder = in_folder
        set_setting('conc_factor', 5)   # TODO: change setting according to data date
        Dialog.__init__(self, parent, "Delay Volume Factor Adjuster", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.X, expand=1)
        self.build_canvas(cframe)
        self.draw_init()

        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT)
        self.build_params_frame(pframe)

        tbframe = Tk.Frame(bframe)
        tbframe.pack(side=Tk.LEFT, fill=Tk.X, expand=1)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tbframe)
        self.toolbar.update()

    def get_canvas_width( self ):
        return int( self.mpl_canvas_widget.cget('width'))

    def build_canvas(self, cframe):
        figsize = (17,6) if is_low_resolution() else (23,8)
        fig = plt.figure(figsize=figsize)
        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        ax1 = fig.add_subplot(131, projection='3d') # it seems this must be done after creation of self.mpl_canvas
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)
        axes = [ax1, ax2, ax3]
        self.fig = fig
        self.axes = axes

        fig.tight_layout()
        fig.subplots_adjust(top=0.88, wspace=0.15, left=0.03, right=0.97)

    def build_params_frame(self, frame):
        width = int(self.mpl_canvas_widget.cget('width'))//3

        space = Tk.Frame(frame, width=width//3)
        space.pack(side=Tk.RIGHT)
        pframe = Tk.Frame(frame)
        pframe.pack(side=Tk.RIGHT)
        space = Tk.Frame(frame, width=width//3)
        space.pack(side=Tk.RIGHT)

        label = Tk.Label(pframe, text="DVF: ")
        label.grid(row=0, column=0)
        label = Tk.Label(pframe, text="%.3g" % self.sm_info.factor)
        label.grid(row=0, column=1)
        space = Tk.Label(pframe, text="   ")
        space.grid(row=0, column=2)
        label = Tk.Label(pframe, text="Conc Scale: ")
        label.grid(row=0, column=3)
        label = Tk.Label(pframe, text="%.3g" % self.sm_info.vscale)
        label.grid(row=0, column=4)

    def show(self):
        self._show(wait=False)
        self.waited_widget = Tk.Frame(self)
        self.wait_window(self.waited_widget)

    def ok(self):
        self.apply()
        self.close()

    def exit(self):
        self.close()

    def close(self):
        self.withdraw()
        self.update_idletasks()
        self.grab_release()
        self.waited_widget.destroy()

    def apply(self):
        self.applied = True

    def draw_init(self):
        xdata = self.xdata

        optimizer = DelayOptimizer(self.in_folder, xdata)
        self.slice_ = slice_ = optimizer.get_slice()
        smt, smx, smy = optimizer.get_initial_data()

        ax1, ax2, ax3 = self.axes
        self.fig.suptitle("Delay Volume Factor Adjustment for " + self.in_folder, fontsize=20)
        xdata.plot(ax1)
        init_sm_info = optimizer.sm_info
        self.draw_elution(ax2, xdata, slice_, smt, smx, smy, sm_info=init_sm_info)

        smt, smx, smy, sm_info = optimizer.get_optimized_data()
        self.draw_elution(ax3, xdata, slice_, smt, smx, smy, sm_info=sm_info, mapped=True)

        y_channel_p = optimizer.get_y_channel_point()
        for ax in [ax2, ax3]:
            ax.plot(*y_channel_p, 'o', color='yellow')

        self.mpl_canvas.draw()
        self.sm_info = sm_info

    def draw_elution(self, ax, xdata, slice_, smt, smx, smy, sm_info=None, mapped=False):
        adjust_text = " (optimized mapping)" if mapped else " (input data)"
        pick_pos = get_xray_picking()
        ax.set_title("Elution at Q=%.2g" % pick_pos + adjust_text, fontsize=16)
        if mapped:
            start = sm_info.start
            stop = len(xdata.e_y)
            e_x = np.arange(start, stop)
            e_y = xdata.e_y[start:stop]
        else:
            e_y = xdata.e_y
            e_x = np.arange(len(e_y))

        ax.plot(e_x, e_y, color='orange')

        if mapped:
            axt = ax
            smy *= sm_info.vscale
            sm_slice = slice(sm_info.start, sm_info.stop)
            x_ = smx[sm_slice]
            y_ = smy[sm_slice]
        else:
            axt = ax.twinx()
            axt.grid(False)
            x_ = smx
            y_ = smy

        axt.plot(x_, y_, color='blue', alpha=0.3)

        if sm_info is not None:
            for k in [sm_info.yc_peak, sm_info.lin_start, sm_info.lin_end]:
                axt.plot(smx[k], smy[k], 'o', color='red')

        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)

        f, t = [slice_.start, slice_.stop]
        for k, p in enumerate([f, t]):
            ax.plot([p, p], [ymin, ymax], ':', color='gray')

        rect = mpl_patches.Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax.add_patch(rect)



"""
    Main.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""

def main():
    from multiprocessing import Process, Queue
    from .SharedInfo import SharedInfo, NUM_QUEUES
    from .Loader import loader
    from .RgComputer import rg_computer
    from .Recognizer import recognizer
    from .GUI import gui_loop
    from .Dummy import dummy

    queues = [Queue() for i in range(NUM_QUEUES)]

    print("main")
    si = SharedInfo()

    loader_proc = Process(target=loader, args=(si, 0, queues))
    loader_proc.start()

    rg_proc = Process(target=rg_computer, args=(si, 1, queues))
    rg_proc.start()

    recog_proc = Process(target=recognizer, args=(si, 2, queues))
    recog_proc.start()

    gui_proc = Process(target=gui_loop, args=(si, 3, queues))
    gui_proc.start()

    if False:
        dummy_proc = Process(target=dummy, args=(si, 9))
        dummy_proc.start()



"""
    RollingStats.py

    Copyright (c) 2020-2023, Masatsuyo Takahashi, KEK-PF
"""

import numpy        as np

"""
    from:
        https://stackoverflow.com/questions/27427618/how-can-i-simply-calculate-the-rolling-moving-variance-of-a-time-series-in-pytho
        https://rigtorp.se/2011/01/01/rolling-statistics-numpy.html
"""
def rolling_window(a, window):
    pad = np.ones(len(a.shape), dtype=int)
    pad[-1] = window-1
    pad = list(zip(pad, np.zeros(len(a.shape), dtype=int)))
    a = np.pad(a, pad,mode='reflect')
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)



"""
    SecSaxs.DataTreatment.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import os
import logging
from collections import OrderedDict
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.Mapping.MappingV2Utils import make_mapped_info_for_v2
from .DataSet import copy_create_dataset_from_sd
class DataTreatment:
    def __init__(self, **kwargs):
        self.logger = logging.getLogger(__name__)
        self._dict = OrderedDict()
        self.path = None
        self.base_curve_info = None
        self.corrected_sd = None
        if len(kwargs) > 0:
            self._dict["route"] = kwargs.pop("route", None)
            self._dict["trimming"] =  kwargs.pop("trimming", 1)
            self._dict["correction"] = kwargs.pop("correction", 0)
            self._dict["unified_baseline_type"] = kwargs.pop("unified_baseline_type", 1)
            assert len(kwargs) == 0

    def save(self, path=None):      
        if path is None:
            from molass_legacy.Optimizer.TheUtils import get_treatment_path
            path = get_treatment_path()
        with open(path, "w") as fh:
            fh.write(self.__repr__())

    def load(self, path=None, optimizer_folder=None):
        from molass_legacy.KekLib.EvalUtils import eval_file
        if path is None:
            from molass_legacy.Optimizer.TheUtils import get_treatment_path
            path = get_treatment_path(optimizer_folder=optimizer_folder)
        self.path = path

        def replacer(code):
            return code.replace("DataTreatment", "OrderedDict")

        self._dict = eval_file(path,
                               locals_=globals(),   # for OrderedDict
                               replacer=replacer)

        unified_baseline_type = self.get("unified_baseline_type")
        set_setting("unified_baseline_type", unified_baseline_type)
        self.logger.info("treatment loaded with unified_baseline_type=%d", unified_baseline_type)

    def keys(self):
        return ["route", "trimming", "correction", "unified_baseline_type"]

    def get(self, key):
        return self._dict.get(key)

    def __repr__(self):
        return str(self._dict).replace("OrderedDict", "DataTreatment")

    def get_trimmed_sd(self, sd, pre_recog, debug=False):
        if debug:
            import Trimming.OptimalTrimming
            from importlib import reload
            reload(Trimming.OptimalTrimming)
        from molass_legacy.Trimming.OptimalTrimming import OptimalTrimming

        mt = OptimalTrimming(sd, pre_recog, debug=debug)
        base_curve_info = mt.get_base_curve_info()

        trimming = self._dict["trimming"]

        if trimming == 1 or True:
            sd_, new_prerecog = sd._get_analysis_copy_impl(pre_recog, return_also_new_prerecog=True, debug=True)
            trimmed = "trimmed"
            # no mt.set_info() to keep the trimming info

        elif trimming == 2:
            trimmed = "minimally-trimmed"
            mt.set_info()
            sd_ = sd._get_analysis_copy_impl(pre_recog)
        else:
            sd_ = sd.get_copy()
            trimmed = "non-trimmed"
            base_curve_info = None

        sd_copy = copy_create_dataset_from_sd(sd, sd_)
    
        if False:
            xr_curve0 = sd.get_xray_curve()
            xr_curve1 = sd_copy.get_xr_curve()
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("get_trimmed_sd")
                ax1.plot(xr_curve0.x, xr_curve0.y)
                ax2.plot(xr_curve1.x, xr_curve1.y)
                fig.tight_layout()
                plt.show()

        self.base_curve_info = base_curve_info
        self.logger.info("got %s data by treatment from %s", trimmed, self.path)

        self.sd_orig = sd
        self.pre_recog = new_prerecog
        self.trimmed_sd = sd_copy

        return sd_copy

    def get_corrected_sd(self, sd, pre_recog, trimmed_sd, debug=False):
        if self._dict["correction"]:
            from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
            x_curve = trimmed_sd.get_xray_curve()
            mapped_info = make_mapped_info_for_v2(x_curve)
            sd_copy = trimmed_sd.copy()
            sd_copy = get_corrected_sd_impl(sd_copy, sd, pre_recog, mapped_info=mapped_info)
            corrected = "corrected"
            self.corrected_sd = sd_copy
        else:
            sd_copy = trimmed_sd
            corrected = "non-corrected"

        self.logger.info("got %s data by treatment", corrected)
        return sd_copy

    def get_treated_sd(self, sd, pre_recog, debug=False):
        trimmed_sd = self.get_trimmed_sd(sd, pre_recog, debug=debug)
        corrected_sd = self.get_corrected_sd(sd, pre_recog, trimmed_sd)
        return corrected_sd

    def get_base_curve_info(self):
        return self.base_curve_info

    def get_base_curve(self):
        if self.base_curve_info is None:
            uv_base_curve = None
        else:
            uv_base_curve = self.base_curve_info[0]
        return uv_base_curve

def demo(in_folder, sd):
    import molass_legacy.KekLib.DebugPlot as plt
    from molass_legacy.DataStructure.MatrixData import simple_plot_3d
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition

    print(in_folder)
    pre_recog = PreliminaryRecognition(sd)
    treat = DataTreatment(route="v2", trimming=2, correction=1)
    trimmed_sd = treat.get_trimmed_sd(sd, pre_recog)
    corrected_sd = treat.get_corrected_sd(sd, pre_recog, trimmed_sd)

    D0, E0, qv0, xr_curve0 = sd.get_xr_data_separate_ly()
    D1, E1, qv1, xr_curve1 = trimmed_sd.get_xr_data_separate_ly()
    D2, E2, qv2, xr_curve2 = corrected_sd.get_xr_data_separate_ly()

    with plt.Dp():
        fig, axes = plt.subplots(ncols=3, figsize=(18,5), subplot_kw=dict(projection="3d"))
        simple_plot_3d(axes[0], D0, x=qv0, y=xr_curve0.x)
        simple_plot_3d(axes[1], D1, x=qv1, y=xr_curve1.x)
        simple_plot_3d(axes[2], D2, x=qv2, y=xr_curve2.x)
        fig.tight_layout()
        plt.show()



"""
    Models/Stochastic/Tripore.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.Tentative import Model
from molass_legacy.Models.ElutionModelUtils import x_from_height_ratio_impl
from molass_legacy.Models.Stochastic.TriporeImpl import tripore_func, guess_multiple_impl

class Tripore(Model):
    def __init__(self, **kwargs):
        super(Tripore, self).__init__(tripore_func, **kwargs)

    def get_name(self):
        return "STC"

    def is_traditional(self):
        return False

    def guess_multiple(self, x, y, num_peaks, debug=False):
        return guess_multiple_impl(x, y, num_peaks, debug=debug)

    def eval(self, params=None, x=None):
        return self.func(x, *params)

    def x_from_height_ratio(self, ecurve, ratio, params):
        return x_from_height_ratio_impl(tripore_func, ecurve, ratio, *params, needs_ymax=True, full_params=True)

    def get_params_string(self, params):
        return 'scale=%g, N=%g, T=%g, rg=%g, x0=%g' % tuple(params)

    def adjust_to_xy(self, params_list, x, y, props=None, devel=False):
        if props is None:
            areas = []
            for p in params_list:
                cy = tripore_func(x, *p)
                areas.append(np.sum(cy))
            props = np.array(areas)/np.sum(areas)

        print("props=", props)

        if devel:
            def plot_with_params_list(title, params_list):
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title(title)
                    ax.plot(x, y)
                    cy_list = []
                    for p in params_list:
                        cy = tripore_func(x, *p)
                        cy_list.append(cy)
                        ax.plot(x, cy, ":")
                    ty = np.sum(cy_list, axis=0)
                    ax.plot(x, ty, ":", color="red")
                    fig.tight_layout
                    plt.show()
            plot_with_params_list("before conversion", params_list)

        params_array = np.array(params_list)

        def objective(p):
            cy_list = []
            areas = []
            for params in p.reshape(params_array.shape):
                cy = tripore_func(x, *params)
                cy_list.append(cy)
                areas.append(np.sum(cy))
            ty = np.sum(cy_list, axis=0)
            props_ = np.array(areas)/np.sum(areas)
            fv = np.sum((ty - y)**2) + np.sum((props_ - props)**2)
            return fv

        ret = minimize(objective, params_array.flatten(), method='Nelder-Mead')
        print("ret.success=", ret.success)
        converted_array = ret.x.reshape(params_array.shape)

        if devel:
            plot_with_params_list("after conversion", converted_array)
        return converted_array

def tripore_test_impl(advanced_frame):
    editor = advanced_frame.editor
    editor_frame = editor.get_current_frame()
    model = editor_frame.model
    print("tripore_test_impl", model.get_name(), model.__class__)
    params_array = editor.get_current_params_array()

    fx = editor_frame.fx
    x = editor_frame.x
    y = editor_frame.y
    uv_y = editor_frame.uv_y
    num_peaks = 3
    guess_multiple_impl(fx, y, num_peaks, debug=True)
    return


    converted_list = model.adjust_to_xy(params_list, fx, uv_y)

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        ax1.plot(x, uv_y, color="blue")
        for params in converted_list:
            cy = tripore_func(fx, *params)
            ax1.plot(x, cy, ":")
        ax2.plot(x, y, color="orange")
        for params in params_list:
            cy = tripore_func(fx, *params)
            ax2.plot(x, cy, ":")
        fig.tight_layout()
        plt.show()



"""

    GuiMain.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF

"""
import sys
import re
import os
import time
import warnings
from molass_legacy.KekLib.EnvironCheck import executables_check
from molass_legacy.KekLib.OurMatplotlib import mpl_1_5_backward_compatible_init, mpl_font_init
mpl_1_5_backward_compatible_init()
mpl_font_init()

from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker, log_exception
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry, is_almost_empty_dir
from molass_legacy.KekLib.ChangeableLogger import Logger
from molass_legacy.KekLib.OurTkinter import Tk, ttk, FileDialog, Font, is_empty_val
from molass_legacy.KekLib.TkUtils import adjusted_geometry, PositionSynchronizer
from molass_legacy.KekLib.TkSupplements import SlimButton, BlinkingFrame
from molass_legacy.KekLib.TkSizableWidgets import SizableLabel, SizableEntry, BlinkingLabel
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry, FileEntry
try:
    import molass_legacy.KekLib.CustomMessageBox as MessageBox
except:
    import molass_legacy.KekLib.OurMessageBox as MessageBox
print( 'import molass_legacy.KekLib.CustomMessageBox ok' )
from molass_legacy.KekLib.RecentFolders import RecentFolders
import Decomposer, DataStructure, Extrapolation, GuinierAnalyzer
from Menus.GuiSettings import GuiSettingsMenu
from Menus.GuiSecTools import GuiSecToolsMenu

from Menus.GuiDenssTools import GuiDenssToolsMenu
from Menus.GuiTutorials import GuiTutorialsMenu
from Menus.GuiReferences import GuiReferencesMenu
from Menus.GuiDevelopment import GuiDevelopmentMenu

from molass_legacy.SerialAnalyzer.SerialDataLoader import SerialDataLoader
from molass_legacy.SerialAnalyzer.SerialDataUtils import get_uv_filename, get_mtd_filename
from molass_legacy.DataStructure.MeasuredData import MeasuredData
from molass_legacy.Trimming.OutlineFigure import OutlineFigure
from molass_legacy.SerialAnalyzer.AbnormalityCheck import exclude_abnormality
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
from molass_legacy.QuickAnalysis.Analyzer import Analyzer
from molass_legacy._MOLASS.Version import get_version_string
from Rank.RankView import RankViewMenu
from GuiParts.UvFolderEntry import UvFolderEntry
from molass_legacy._MOLASS.SerialSettings import ( initialize_settings,
                                    get_setting, set_setting, save_settings,
                                    clear_temporary_settings, restore_default_setting,
                                    get_settings_folder )
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting, set_dev_setting
import AutorgKek.Settings

analysis_name_re    = re.compile( r'^(.*?)(\d+)(.*)$' )
distance_symbol_re  = re.compile( r'^(\S+)\s' )

label_bg_color = 'gray40'
label_fg_color = 'white'
label_gb_color_var = 'white smoke'
relief_var = Tk.GROOVE
radio_button_text_length = 20

angular_unit_guide = "; this unit is used only for display purpose."

class GuiMain( Tk.Toplevel ):
    def __init__(self, parent, pid=None, debug=0):
        from molass_legacy._MOLASS.Debug import DebugInfo
        dpi_aware = False
        self.debug_info = DebugInfo(debug=debug)
        self.clear_test_mode()

        self.parent_pid = pid
        self.cleaner = None

        """
            dpi_aware is for compatibility with V2
            and not used
        """

        self.testing = False
        if not executables_check(parent):
            parent.after(100, parent.quit)
            return

        initialize_settings()

        self.single_instance = None     # this is required in case of exiting within the following check
        from molass_legacy._MOLASS.AppSingleInstance import single_instance_check
        app_lock_path = get_settings_folder() + '\\lock'
        self.single_instance = single_instance_check(app_lock_path)

        self.default_font = Font.nametofont("TkDefaultFont")
        self.hiresolution   = get_dev_setting( 'hiresolution' )
        self.high_dpi = False

        if self.hiresolution:
            self.default_font.configure(size=16)
            parent.option_add("*Font", self.default_font)
            self.slimbutton = False
            fixed_fontsize  = 16
        else:
            self.slimbutton = True
            fixed_fontsize  = 9

        self.fixed_font = Font.Font( family="Courier", size=fixed_fontsize )

        warnings.filterwarnings("ignore")

        parent.wm_title( get_version_string() + " pid:%d" % os.getpid() )
        parent.report_callback_exception = self.report_callback_exception

        if '-i' in sys.argv:
            from molass_legacy._MOLASS.SerialSettings   import clear_settings
            clear_settings()

        self.parent = parent
        self.tmp_logger = Logger()  # this logger logs into string stream
        clear_temporary_settings()
        self.serial_data = None
        self.measured_data = None

        self.loader     = SerialDataLoader(dialog=self)
        self.analyzer   = Analyzer( self, self.loader )
        self.intial_geometry = None
        self.use_mtd_conc = False
        self.is_executing   = False
        self.dataset_is_ready = False
        self.preproc_dialog = None
        self.keeping_setting_items = None

        Tk.Toplevel.__init__( self, parent )
        self.withdraw()

        if get_setting('enable_debug_plot'):
            from molass_legacy.KekLib.DebugPlot import set_plot_env
            set_plot_env(self)

        self.build_window()
        if '--close_immediately' in sys.argv:
            print( 'closing immediately' )
            self.quit( immediately=True )

        self.after(100, self.init_proc)
        self.pos_sync = PositionSynchronizer(self, self.parent)

    def __del__(self):
        if self.single_instance is not None:
            self.single_instance.clean_up()

    def set_test_mode(self, tester_info):
        self.testing = True
        self.tester_info = tester_info
        set_setting("test_pattern", tester_info.test_pattern)

    def clear_test_mode(self):
        self.testing = False
        self.tester_info = None
        set_setting("test_pattern", None)

    def init_proc(self):
        # initial proc done after GuiMain.__init__
        for k in range(2):
            try:
                self.check_environment()
                self.make_it_more_clever()
                break
            except:
                # ocurred once when D&D in a not ready state
                pass
            time.sleep(1)

        self.after(1000, self.update_menu_states)   # to ensure DENSS Tools menu states update

    def check_environment( self ):
        from molass_legacy.Env.EnvInfo import EnvInfo, set_global_env_info
        env_info = EnvInfo()
        env_info.show_and_log_if_unavailable(self, self.tmp_logger)
        self.env_info = env_info
        set_global_env_info(env_info)

    def make_it_more_clever(self):
        from molass_legacy.SerialAnalyzer.MakeItMoreClever import increase_menu_availability
        increase_menu_availability(self)

    def is_busy( self, auto_apply=True ):
        if auto_apply:
            try:
                if self.analyzer.cancel_reports:
                    # this case is for testing only
                    dialog = self.analyzer.dialog
                    if dialog is None:
                        pass
                else:
                    dialog = self.analyzer.progress_dialog
                    if dialog is None:
                        pass
                    else:
                        text = dialog.button.cget('text')
                        if text == 'OK':
                            self.analyzer.progress_dialog.button.invoke()
            except:
                etb = ExceptionTracebacker()
                print(etb)
                pass
        return self.is_executing

    def report_callback_exception(self, exc, val, tb):
        # This method is to override the Tk method to be able
        # to report to the spawned console in windows application mode.

        if self.tmp_logger is None:
            self.tmp_logger = Logger()
        logger = self.tmp_logger

        etb = ExceptionTracebacker()
        msg = ( 'Overridden report_callback_exception: ' + str( etb )
                + '\n---- %s ----\n' % get_version_string(cpuid=True)
                + 'This message is saved either to "%s"\n' % logger.get_final_log_path()
                + 'or to "molass.log" in the analysis report folder.'  )

        logger.error( msg )
        self.tmp_logger = logger = Logger()     # this line invokes Logger's destructor
                                                # and recovers another logger in case
                                                # the execution continues
        """
            this is necessary for
            some modules such as scipy.interpolate seem to exit
            without giving any chance to the destructor
        """

        if not self.analyzer.logger_visible:
            try:
                MessageBox.showerror( "Tkinter Error", msg, parent=self.parent )
            except:
                # _tkinter.TclError: bad window path name ".!guimain"
                pass

    def adjust_geometry( self ):
        self.update()
        self.deiconify()
        if self.intial_geometry is None:
            self.geometry( adjusted_geometry( self.geometry() ) )
            self.intial_geometry = self.geometry()

    def build_window( self ):
        from GuiParts.FileInfoTable import FileInfoTable    # must be imported after initialize_settings

        global radio_button_text_length

        lib_dir = os.path.dirname( os.path.dirname( os.path.abspath( __file__ ) ) )
        iconpath = lib_dir + '/_MOLASS/molamola.ico'
        self.iconbitmap(default=iconpath)           # this default argument key is for Windows only

        # TODO: recover the previous state
        # main_geometry = get_setting( 'main_geometry' )
        # print( main_geometry )
        # self.geometry( main_geometry )
        # self.geometry( default_main_geometry )

        # temp_label = DynamicLabel(self)
        # fixed_font = temp_label.get_fixed_font()
        # fixed_font = Font.Font(family="FixedSys", size=8 )
        # fixed_font = Font.Font( family="ＭＳ ゴシック", size=10 )
        # fixed_font = Font.Font( family="Lucida Console", size=10 )
        # fixed_font = Font.Font( family="DotumChe", size=10 )
        # fixed_font = Font.Font( family="Consolas", size=10 )
        self.fixed_font = Font.Font( family="Courier", size=9 )

        self.create_menus()

        # basic Layout
        top_space = Tk.Frame( self, height=20 )
        top_space.pack()

        upper_frame = Tk.Frame( self )
        upper_frame.pack( fill=Tk.BOTH, padx=20 )

        lower_frame = Tk.Frame( self )
        lower_frame.pack( fill=Tk.BOTH, expand=1, padx=20 )

        # layout params
        section_label_width = 20
        button_width = 12
        state_show_width = 23
        state_show_sizable_width    = 164
        state_show_sizable_height   = 20
        folder_entry_width          = 60

        grid_row = -1

        # Input Information --------------------------------------------------------
        grid_row += 1
        column_adjust  = Tk.Frame( upper_frame, width=160, height=0 )
        column_adjust.grid( row=grid_row, column=0 )

        input_label  = Tk.Label( upper_frame, text= 'Input', width=section_label_width,
                                    relief=Tk.FLAT, fg=label_fg_color, bg=label_bg_color )
        input_label.grid( row=grid_row, column=0, sticky=Tk.W )

        grid_row += 1
        in_folder_label  = Tk.Label( upper_frame, text= 'Xray Scattering Data Folder: ' )
        in_folder_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.in_folder = Tk.StringVar()
        self.in_folder.set( get_setting( 'in_folder' ) )
        self.in_folder_entry = FolderEntry( upper_frame, textvariable=self.in_folder, width=folder_entry_width,
                                            slimbutton=self.slimbutton,
                                            on_entry_cb=self.on_entry_in_folder )
        self.in_folder_entry.grid( row=grid_row, column=1, sticky=Tk.W )
        self.previous_in_folder = None
        self.previous_uv_file = None
        self.uv_file_info_changed = False
        self.cleared_uv_folder  = False

        file_extension_label  = Tk.Label( upper_frame, text= '  File Ext.: ' )
        file_extension_label.grid( row=grid_row, column=2, sticky=Tk.E )
        self.file_extension  = Tk.StringVar()
        self.file_extension.set( get_setting( 'file_extension' ) )
        file_extension_box = ttk.Combobox( upper_frame, textvariable=self.file_extension,
                                            width=5, justify=Tk.CENTER )
        file_extension_box[ 'values' ] = [ '*.dat', '*.csv', '*.int', '*.*' ]
        file_extension_box.grid( row=grid_row, column=3, sticky=Tk.W )

        self.file_extension.trace( "w", self.file_extension_tracer )

        grid_row += 1
        uv_folder_label  = Tk.Label( upper_frame, text= 'UV Absorbance Data Folder: ' )
        uv_folder_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.uv_folder = Tk.StringVar()
        self.uv_folder.set( get_setting( 'uv_folder' ) )
        self.uv_folder_entry = UvFolderEntry( upper_frame, textvariable=self.uv_folder, width=folder_entry_width,
                                            slimbutton=self.slimbutton,
                                            on_entry_cb=self.on_entry_uv_folder )
        self.uv_folder_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        uv_file_name_label  = Tk.Label( upper_frame, text= '  File Name: ' )
        uv_file_name_label.grid( row=grid_row, column=2, sticky=Tk.E )
        self.uv_file = Tk.StringVar()

        uv_file_frame = Tk.Frame( upper_frame )
        uv_file_frame.grid( row=grid_row, column=3, sticky=Tk.W )

        # TODO: refactoring with FileEntry
        self.uv_file_entry = Tk.Entry( uv_file_frame, textvariable=self.uv_file, width=30 )
        self.uv_file_entry.grid( row=0, column=0, sticky=Tk.W )
        if self.slimbutton:
            b3 = SlimButton( uv_file_frame, text='...', command=self.select_uv_file, width=18, height=22 )
        else:
            b3 = Tk.Button( uv_file_frame, text='...', command=self.select_uv_file )

        b3.grid( row=0, column=1, sticky=Tk.W )

        # DND bind
        self.uv_file_entry.register_drop_target("*")

        def dnd_handler( event ):
            self.select_uv_file(path=event.data)

        self.uv_file_entry.bind("<<Drop>>", dnd_handler)

        # disable UV data
        grid_row += 1
        self.disable_uv_data = Tk.IntVar()
        self.disable_uv_data.set(get_setting('disable_uv_data'))
        self.disable_uv_data_cb = Tk.Checkbutton( upper_frame, variable=self.disable_uv_data,
                                text='disable UV data', state=Tk.DISABLED )
        self.disable_uv_data_cb.grid(row=grid_row, column=2, columnspan=2, sticky=Tk.W, padx=5)
        self.disable_uv_data_tracing = True
        self.disable_uv_data_tracer_busy = False
        self.disable_uv_data.trace('w', self.disable_uv_data_tracer)

        # Conc. Factors
        from GuiParts.ConcFactorsEntry import ConcFactorsEntry

        grid_row += 1
        self.cfs_entry = ConcFactorsEntry(upper_frame, grid_row)

        # pad
        grid_row += 2   # ConcFactorsEntry includes 2 rows
        """
        self.no_absorbance = Tk.IntVar()
        cb = Tk.Checkbutton( upper_frame, text='Use without Absorbance Data',
                    variable=self.no_absorbance )
        cb.grid( row=grid_row, column=1, sticky=Tk.W )
        """

        pad_frame = Tk.Label( upper_frame )
        pad_frame.grid( row=grid_row, column=0 )

        # Analysis Result Setting ----------------------------------------------
        grid_row += 1

        analysis_label  = Tk.Label( upper_frame, text= 'Output', width=section_label_width,
                                    relief=Tk.FLAT, fg=label_fg_color, bg=label_bg_color )
        analysis_label.grid( row=grid_row, column=0, sticky=Tk.W )

        grid_row += 1
        an_folder_label  = Tk.Label( upper_frame, text= 'Analysis Result Folder: ' )
        an_folder_label.grid( row=grid_row, column=0, sticky=Tk.E )
        self.an_folder = Tk.StringVar()
        self.an_folder.set( get_setting( 'an_folder' ) )
        self.an_folder_frame = BlinkingFrame( upper_frame )
        self.an_folder_frame.grid( row=grid_row, column=1, sticky=Tk.W )
        self.an_folder_entry = FolderEntry( self.an_folder_frame, textvariable=self.an_folder, width=folder_entry_width,
                                            slimbutton=self.slimbutton,
                                            on_entry_cb=self.on_entry_an_folder )
        self.an_folder_entry.pack()
        self.an_folder_entry.bind('<FocusOut>', self.an_folder_focusout)
        self.an_folder_frame.objects = [self.an_folder_entry.entry]

        grid_row += 1
        analysis_name_label = Tk.Label( upper_frame, text='Subfolder: ' )
        analysis_name_label.grid( row=grid_row, column=0, sticky=Tk.E )
        analysis_frame = Tk.Frame( upper_frame )
        analysis_frame.grid( row=grid_row, column=1, sticky=Tk.W )
        self.analysis_name = Tk.StringVar()
        self.analysis_name.set( get_setting( 'analysis_name' ) )
        self.analysis_name_entry = Tk.Entry( analysis_frame, textvariable=self.analysis_name, width=20 )
        self.analysis_name_entry.grid( row=0, column=0, sticky=Tk.W )
        self.auto_number = Tk.IntVar()
        auto_number_ = get_setting( 'auto_number' )
        self.auto_number.set( auto_number_ )
        if False:
            self.auto_numbering_cb = Tk.Checkbutton( analysis_frame, variable=self.auto_number,
                                                text='Auto-number (un-check to reuse)' )
                                                # text='Auto-number (un-check this to reuse previous results)' )
            self.auto_numbering_cb.grid( row=0, column=1, sticky=Tk.W, padx=10 )

        if auto_number_ == 1 and not is_empty_val( self.an_folder.get() ):
            analysis_name, _ = self.make_analysis_folder( format_check_only=True )
            self.analysis_name.set( analysis_name )

        self.analysis_name_entry.bind( '<FocusOut>',  self.analysis_name_entry_focus_out )

        grid_row += 1
        result_book_label = Tk.Label( upper_frame, text='Book Name: ' )
        result_book_label.grid( row=grid_row, column=0, sticky=Tk.E )
        self.result_book = Tk.StringVar()
        result_book = get_setting( 'result_book' )
        self.result_book.set( result_book )
        self.result_book_entry = Tk.Entry( upper_frame, textvariable=self.result_book, width=20 )
        self.result_book_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        grid_row += 1
        pad_frame = Tk.Label( upper_frame )
        pad_frame.grid( row=grid_row, column=0 )

        # File Information Table -----------------------------------------------
        grid_row += 1
        grid_row_for_analysis_button = grid_row
        table_label  = Tk.Label( upper_frame, text= 'Xray Scattering Data Files', width=section_label_width,
                                    relief=Tk.FLAT, fg=label_fg_color, bg=label_bg_color )
        table_label.grid( row=grid_row, column=0, sticky=Tk.W )

        fig_table_frame = Tk.Frame( lower_frame )
        fig_table_frame.pack(fill=Tk.X, expand=1, pady=10)

        # navigation button frame
        nav_frame = Tk.Frame(fig_table_frame)
        nav_frame.pack( side=Tk.BOTTOM, fill=Tk.X, expand=1, pady=10)

        self.message_frame = Tk.Frame( fig_table_frame )
        self.message_frame.pack( side=Tk.BOTTOM, fill=Tk.X, expand=1)   # Tk.BOTTOM required!

        table_height = 150 if self.high_dpi else 100
        self.file_info_table = FileInfoTable( fig_table_frame, self, self.loader, self.fixed_font, height=table_height )
        self.file_info_table.pack( side=Tk.RIGHT, fill=Tk.BOTH, expand=1, anchor=Tk.N)

        figsize = (6,8) if self.high_dpi else (4,5)
        self.fig_frame = OutlineFigure(fig_table_frame, self, file_info_table=self.file_info_table, figsize=figsize)
        self.fig_frame.pack(side=Tk.LEFT, fill=Tk.BOTH)

        # fill=Tk.BOTH would make the Scrollbar too long in some cases

        self.btn_frame = Tk.Frame(upper_frame)
        self.btn_frame.grid(row=grid_row, column=1, columnspan=3, sticky=Tk.E)

        self.refresh_button = Tk.Button( self.btn_frame, text='Refresh', command=self.refresh )
        self.refresh_button.grid(row=0, column=0)
        self.presync_button = Tk.Button( self.btn_frame, text='3D View', command=self.show_threedim_dialog )
        self.presync_button.grid(row=0, column=1, padx=10)
        self.restrict_button = Tk.Button( self.btn_frame, text='Data Range', command=self.show_datarange_dialog )
        self.restrict_button.grid(row=0, column=2)
        self.rankview_menu = RankViewMenu(self.btn_frame, self)
        self.rankview_menu.grid(row=0, column=3, padx=10)
        self.baseline_button = Tk.Button( self.btn_frame, text='Baseline', command=self.show_lpm_inspector )
        self.baseline_button.grid(row=0, column=4)

        # Analysis start button ---------------------------------------------------
        nav_btn_frame = Tk.Frame(nav_frame)
        nav_btn_frame.pack(side=Tk.RIGHT)

        exit_button = Tk.Button(nav_frame, text="▼ Exit", command=self.quit)
        exit_button.pack(side=Tk.LEFT, padx=10)

        button_frame = Tk.Frame( nav_btn_frame )
        button_frame.pack()

        self.analysis_button = Tk.Button( button_frame, text="▶ Analysis start", command=self.analyze_all )
        self.analysis_button.grid( row=0, column=0, padx=5 )
        self.fully_automatic  = Tk.IntVar()
        self.fully_automatic.set(get_setting('fully_automatic'))
        self.fully_automatic_cb = Tk.Checkbutton( button_frame, text="fully automatic", variable=self.fully_automatic )
        self.fully_automatic_cb.grid( row=0, column=1, padx=5 )
        self.buttons = [self.refresh_button, self.restrict_button, self.presync_button, self.rankview_menu, self.baseline_button,
                            self.analysis_button, self.fully_automatic_cb,
                            self.disable_uv_data_cb, self.cfs_entry,
                            exit_button]

        self.enable_synthesized_lrf = get_setting('enable_synthesized_lrf')
        if self.enable_synthesized_lrf:
            self.synthesized_lrf  = Tk.IntVar()
            self.synthesized_lrf.set(get_setting('synthesized_lrf'))
            self.synthesized_lrf_cb = Tk.Checkbutton( button_frame, text="synthesized LRF", variable=self.synthesized_lrf )
            self.synthesized_lrf_cb.grid( row=0, column=2, padx=5 )
            self.buttons.append(self.synthesized_lrf_cb)

        self.update_button_states()
        self.update_analysis_button_state()

        # Plot results button ---------------------------------------------------
        self.plot_button = Tk.Button( button_frame, text="Plot Results", command=self.plot_results )
        self.plot_button.grid( row=0, column=1, padx=5 )
        self.update_plot_button_state()

        self.in_folder_has_been_changed = False
        self.settings_have_been_changed = False

        # it seemd that '<Control-c>' should be bound from the top-level.
        self.bind( '<Control-c>', self._on_control_c )
        self.bind( '<Button-1>', self._on_click )
        # self.bind( '<Return>', self._on_enter )
        # self.bind( '<Control-a>', self._on_enter )
        self.bind( '<Escape>', self._on_escape )
        self.in_folder_entry.bind( '<FocusIn>', self.on_in_folder_entry_focus_in )
        self.in_folder_entry.bind( '<FocusOut>',  self.on_in_folder_entry_focus_out )

        # message frame
        self.guide_message = Tk.Label(self.message_frame, bg='white')
        self.guide_message.pack(fill=Tk.X, pady=10)

        self.set_state_guide_message("Specify input/output folders")

        # getting ready to show
        self.adjust_geometry()

        invoke_on_entry = False
        if not is_empty_val( self.in_folder.get() ):
            invoke_on_entry = self.ask_resume()

        # file_info_table does not grow in size after geometry setting
        self.file_info_table.refresh()

        if invoke_on_entry:
            self.on_entry_in_folder()

        uv_folder_ = self.uv_folder.get()
        if not is_empty_val( uv_folder_ ):
            if is_empty_val( self.uv_file.get() ):
                uv_file = get_uv_filename( uv_folder_ )
                self.uv_file.set( uv_file )

        self.datarange_dialog = None
        self.protocol( "WM_DELETE_WINDOW", self.quit )

    def ask_resume(self):
        self.update()
        ret = MessageBox.askyesno("Reload Confirmation",
            'Previous input was from\n'
            '%s\n'
            'Would you like to have it loaded?\n'
            % self.in_folder.get(), parent=self,
            )
        if ret:
            invoke_on_entry = True
        else:
            self.in_folder.set(None)
            set_setting("in_folder", None)
            invoke_on_entry = False
        return invoke_on_entry

    def set_state_guide_message(self, guide_message, fg='black'):
        self.message_frame.update()
        try:
            self.guide_message.config(text=guide_message.replace('\n', ' '), fg=fg)
        except:
            # _tkinter.TclError: invalid command name ".!guimain.!frame3.!frame.!frame2.!label"
            log_exception(self.tmp_logger, "self.guide_message.config(...): ")

    def file_extension_tracer( self, *args ):
        set_setting( 'file_extension', self.file_extension.get() )
        self.on_entry_in_folder()

    def refresh( self, keeping_setting_items=None):
        print( 'refresh' )
        self.keeping_setting_items = keeping_setting_items
        self.previous_in_folder = None  # this will induce clear_temporary_settings()
        self.in_folder_entry.on_entry(refresh=True)
        # self.uv_folder_entry.on_entry()

        if self.pre_recog.restrict_info_changed():
            yn = MessageBox.askyesno("Question",
                            "Data Range info seems to have been changed mannually.\n"
                            "Do you wish reset it to the default?",
                            parent=self)
            if yn:
                self.pre_recog.reset_restrict_info()

    def show_datarange_dialog( self ):
        from molass_legacy.Trimming import DataRangeDialog
        range_editor_info = get_setting('range_editor_info')
        decomp_editor_info = get_setting('decomp_editor_info')

        if range_editor_info is not None or decomp_editor_info is not None:
            ok = MessageBox.askyesno( 
                            'Confirmation Question',
                            'Previously memorized analysis range info exists.\n'
                            'Applying a new data restriction will reset that info to default.\n'
                            'Are you sure to proceed?',
                            parent=self
                            )
            if not ok:
                return

        self.datarange_dialog = DataRangeDialog(self.parent, self.pre_recog)
        self.datarange_dialog.show()

        self.draw_figure()

    def has_datarange_dialog(self):
        return self.datarange_dialog is not None

    def create_menus( self ):
        import _MOLASS.SerialSettings
        menubar = Tk.Menu( self )
        self.config( menu=menubar )
        self.menubar = menubar

        menu0 = Tk.Menu( menubar, tearoff=0 )
        self.recent_menu = Tk.Menu( menubar, tearoff=0 )

        menubar.add_cascade( label="Folder", menu=menu0 )
        menu0.add_cascade( label="Recently used data folders", menu=self.recent_menu )
        self.recent_folders = RecentFolders( size=20, settings=_MOLASS.SerialSettings  )
        self.recent_menu_add()
        menu0.add_command( label="Exit", command=self.quit )

        menu1 = GuiSettingsMenu( self, menubar )
        self.menu3 = GuiSecToolsMenu( self, menubar )
        self.menu4 = GuiDenssToolsMenu( self, menubar )
        self.menu5 = GuiTutorialsMenu( self, menubar )
        menu6 = GuiReferencesMenu( self, menubar )
        menu7 = GuiDevelopmentMenu( self, menubar )

    def recent_menu_add( self ):
        for folder in self.recent_folders.get_sorted_list():
            path = folder[1]
            # print( 'add recent menu', path )
            self.recent_menu.add_command( label=path, command=lambda path_=path: self.set_in_folder( path_ ) )

        self.update()

    def recent_menu_delete( self ):
        for folder in self.recent_folders.get_sorted_list():
            path = folder[1]
            # print( 'deleting menu ', path )
            try:
                self.recent_menu.delete( path )
            except:
                print( 'error deleting menu ', path )

    def askdirectory( self, entry_variable ):
        entered_path = entry_variable.get()
        # print 'entered_path=', entered_path
        dir_ = os.path.dirname( entered_path ).replace( '/', '\\' )
        # print 'dir_=', dir_
        f = FileDialog.askdirectory( initialdir=dir_, parent=self )
        return f

    def set_in_folder( self, path ):
        # print( 'set_in_folder', path )
        self.in_folder.set( path )
        self.on_entry_in_folder()
        self.settings_have_been_changed = True

    def set_uv_folder( self, path, filename ):
        # print( 'set_uv_folder', path )
        self.uv_folder.set( path )
        self.on_entry_uv_folder( filename=filename )
        self.settings_have_been_changed = True

    def select_uv_file(self, path=None):
        if path is None:
            f = FileDialog.askopenfilename( initialdir=self.uv_folder.get(), parent=self )
            if not f:
                return
        else:
            f = path

        dir_, file = os.path.split( f )
        self.on_entry_uv_file( dir_, file  )
        self.settings_have_been_changed = True

    def update_button_states(self):
        state = Tk.NORMAL if (self.dataset_is_ready and not self.is_executing) else Tk.DISABLED
        for k, btn in enumerate(self.buttons):
            btn.config(state=state)

        disable_uv_data = get_setting('disable_uv_data')
        if not disable_uv_data:
            if is_empty_val(self.uv_file.get()):
                for w in [self.cfs_entry, self.disable_uv_data_cb]:
                    w.config(state=Tk.DISABLED)

        self.update_menu_states()
        self.update()

    def update_menu_states(self):
        self.menu3.update_states()
        self.menu4.update_states()
        self.menu5.update_states()

    def update_analysis_button_state( self ):
        if self.is_executing:
            state_ = Tk.DISABLED
        else:
            if self.dataset_is_ready:
                num_data_rows = self.file_info_table.table.cells_array.shape[0] - 1
                use_xray_conc = get_setting( 'use_xray_conc' ) == 1
                use_mtd_conc = get_setting( 'use_mtd_conc' ) == 1
                if use_xray_conc or use_mtd_conc:
                    state_ = Tk.NORMAL if num_data_rows > 0 else Tk.DISABLED
                else:
                    state_ = Tk.NORMAL if num_data_rows > 0 and self.uv_file_exists else Tk.DISABLED
            else:
                state_ = Tk.DISABLED
        try:
            self.analysis_button.config( state=state_ )
            self.update()
        except:
            # _tkinter.TclError: invalid command name ".!guimain.!frame3.!frame.!frame.!frame.!frame.!button"
            pass

    def update_all_button_states(self):
        self.update()
        self.update_button_states()
        self.update_analysis_button_state()
        self.update_plot_button_state()
        self.update()

    def on_entry_in_folder( self, refresh=False ):
        # refresh is currently not used

        self.uv_file_info_changed = False
        self.dataset_is_ready = False
        self.config( cursor='wait' )
        self.update_all_button_states()
        self.uv_file_exists = False

        self.conc_serie = None
        in_folder_ = self.in_folder.get()

        if in_folder_ != self.previous_in_folder:
            kept_items = None
            if refresh and self.keeping_setting_items is not None:
                kept_items = []
                for item in self.keeping_setting_items:
                    kept_items.append(get_setting(item))
            self.on_in_folder_change()
            self.load_start_time = time.time()
            if kept_items is not None:
                for item, value in zip(self.keeping_setting_items, kept_items):
                    set_setting(item, value)
            self.serial_data = None
            self.measured_data = None
            self.previous_in_folder = in_folder_
            self.cleared_uv_folder  = False

        self.recent_menu_delete()
        self.recent_folders.add( in_folder_ )
        self.recent_menu_add()

        set_setting( 'in_folder', in_folder_ )
        self.file_info_table.refresh()
        self.update()

        self.config( cursor='' )

        if self.file_info_table.num_rows == 0:
            self.set_in_folder_error()
            self.update()   # necessary for the file extension combobox to shrink
            MessageBox.showerror( "Xray Scattering Data Folder Error", "No Data File in '%s'" % in_folder_, parent=self )
            self.set_in_folder_error()  # call it again to focus
            return

        # need to restore fg if the case is after the error above
        self.in_folder_entry.config( fg='black' )

        self.loader.reset_input_status()
        if self.tmp_logger is None:
            self.tmp_logger = Logger()  # this logger logs into string stream
        else:
            # self.tmp_logger should have been created after do_analysis
            pass

        if self.parent_pid is not None:
            from molass_legacy.KekLib.CleanerThread import CleanerThread
            with_parent_pid = " with parent_pid=%d" % self.parent_pid
            self.cleaner = CleanerThread(self.parent_pid)
        else:
            with_parent_pid = ""
            self.cleaner = None
        self.tmp_logger.info('logging started on_entry_in_folder in %s%s', get_version_string(cpuid=True), with_parent_pid)

        if self.uv_file_info_changed:
            # keep the current uv file info
            self.loader.reset_current_status()
            self.after(500, self.prepare_serial_data)
        else:
            filename = get_uv_filename( in_folder_ )
            if filename is None:
                filename = get_mtd_filename( in_folder_ )
                if filename is None:
                    if not self.cleared_uv_folder:
                        # don't do this for the same in_folder_
                        self.uv_folder.set( '' )
                        self.uv_file.set( '' )
                        set_setting( 'uv_folder',   None )  # clear
                        set_setting( 'uv_file',     None )  # clear
                        self.cleared_uv_folder = True
                        set_setting( 'use_xray_conc', 1 )
                        from molass_legacy._MOLASS.SerialSettings import do_xray_conc_temporary_settings
                        do_xray_conc_temporary_settings()
                        self.after(500, self.prepare_serial_data)
                else:
                    # 
                    folder, file = os.path.split(filename)
                    self.uv_folder.set( folder )
                    self.uv_file.set( file )

                    set_setting( 'use_xray_conc', 0 )
                    set_setting( 'use_mtd_conc', 1 )
                    self.use_mtd_conc = True
                    set_setting( 'mtd_file_path', filename )
                    # self.fully_automatic.set(0)
                    from molass_legacy._MOLASS.SerialSettings import do_microfluidic_temporary_settings
                    do_microfluidic_temporary_settings()
                    self.after(500, self.prepare_serial_data)
            else:
                self.set_uv_folder( in_folder_, filename )
                set_setting( 'use_xray_conc', 0 )
                # verify the need to call self.prepare_serial_data
                # self.after(500, self.prepare_serial_data)

        self.lacking_q_values_warned = False
        self.fig_frame.clear_figures()
        self.set_state_guide_message("Waiting for the completion\nof data loading.")
        self.update()
        # update states
        self.update_uv_folder_entry_state()
        self.settings_have_been_changed = True

    def clear_entries(self):
        self.cfs_entry.reset_entries()
        clear_temporary_settings()
        self.update()

    def on_in_folder_change(self):
        self.clear_entries()

        if self.disable_uv_data_tracer_busy:
            """
            do nothing when called during the tracer process.
            """
            pass
        else:
            """
            this is the case when the input folder has been changed
            after disabling the previous folder.
            """
            self.disable_uv_data_tracing = False
            self.disable_uv_data.set(get_setting('disable_uv_data'))
            self.update()
            self.disable_uv_data_tracing = True

    def set_in_folder_error( self ):
        self.in_folder_entry.config( fg='red' )
        self.in_folder_entry.focus_force()

    def on_entry_uv_folder( self, filename=None ):
        if self.loader.is_busy:
            # TODO : better to get rid of this case
            if self.tmp_logger is not None:
                self.tmp_logger.info('on_entry_uv_folder is skipped because the loader is busy.')
                return

        uv_folder_ = self.uv_folder.get()

        if filename is None:
            filename = get_uv_filename( uv_folder_ )
            if filename is None:
                self.set_uv_folder_error()
                MessageBox.showerror( "Absorbance Folder Error", "No absorbance file in '%s'." % uv_folder_, parent=self )
                self.uv_file.set( "<Not found>" )
                return

        self.uv_file_info_changed = True
        self.uv_file.set( filename )
        set_setting( 'uv_folder', uv_folder_ )
        set_setting( 'uv_file', filename )
        self.uv_file_exists = True

        data_folder = get_setting( 'in_folder' ).replace( '\\', '/' )
        conc_folder = get_setting( 'uv_folder' ).replace( '\\', '/' )
        conc_file   = get_setting( 'uv_file'   )
        if conc_file.find( '*' ) >= 0:
            conc_file = None

        # self.analysis_button.config( state=Tk.NORMAL )

        if not is_empty_val(self.previous_uv_file):
            self.set_state_guide_message("Re-loading the data")

        self.previous_uv_file = filename
        self.update_analysis_button_state()
        self.loader.load_from_folders( data_folder, uv_folder=conc_folder, uv_file=conc_file )
        # serial_data construction will be delayed until self.analyze_all
        # to keep this thread running without waiting here

        self.settings_have_been_changed = True
        self.after(500, self.prepare_serial_data)

    def set_uv_folder_error( self ):
        self.uv_folder_entry.config( fg='red' )

    def on_entry_uv_file( self, dir_, file ):
        self.uv_file.set( file )
        self.uv_folder.set( dir_ )

        if not self.uv_folder_entry.check():
            return

        if not self.uv_file_check():
            return

        self.set_uv_folder( dir_, file )

    def uv_file_check( self ):
        # TODO: include these checks in a custom widget class
        ok_ = False
        uv_file = self.uv_file.get()
        uv_folder = self.uv_folder.get()
        if uv_file.find( '*' ) >= 0:
            file_ = get_uv_filename(uv_folder, glob=uv_file)
            if file_ is None:
                self.set_uv_file_error()
                MessageBox.showerror( "Absorbance File Error", "No Absorbance File in '%s'" % uv_folder, parent=self )
            else:
                self.uv_file_entry.config( fg='black' )
                ok_ = True
        else:
            uv_filepath = os.path.join( uv_folder, uv_file );
            if is_empty_val( uv_file ):
                self.set_uv_file_error()
                MessageBox.showerror( "Absorbance File Error", "The Absorbance File is required.", parent=self )
            elif not os.path.exists( uv_filepath ):
                self.set_uv_file_error()
                MessageBox.showerror( "Absorbance File Error", "'%s' does not exist." % uv_file, parent=self )
            elif not os.path.isfile( uv_filepath ):
                self.set_uv_file_error()
                MessageBox.showerror( "Absorbance File Error", "'%s' is not a file." % uv_file, parent=self )
            else:
                self.uv_file_entry.config( fg='black' )
                ok_ = True

        if not ok_:
            self.uv_file_entry.focus_force()

        return ok_

    def set_uv_file_error( self ):
        self.uv_file_entry.config( fg='red' )

    def on_in_folder_entry_focus_in( self, event ):
        self.in_folder_ext_state = self.in_folder.get() + self.file_extension.get()

    def on_in_folder_entry_focus_out( self, event ):
        # print( 'on_in_folder_entry_focus_out' )
        in_folder = self.in_folder.get()
        file_extension = self.file_extension.get()
        in_folder_ext_state = in_folder + file_extension
        if in_folder_ext_state == self.in_folder_ext_state:
            return

        uv_folder = self.uv_folder.get()
        if is_empty_val( in_folder ) or not is_empty_val( uv_folder ):
            return

        self.on_entry_in_folder()

    def on_entry_an_folder( self ):
        # print( 'on_entry_an_folder' )
        self.an_folder_frame.stop()
        self.update()

        if not self.an_folder_check():
            return

        # self.an_folder_entry.config(fg=None)
        self.config( cursor='wait' )    # doesn't seem to work
        self.update()
        an_folder_ = self.an_folder.get()
        set_setting( 'an_folder', an_folder_ )
        self.settings_have_been_changed = True
        self.update_plot_button_state()
        if self.serial_data is not None:
            self.draw_figure()
        self.config( cursor='' )
        self.make_it_more_clever()

    def an_folder_focusout( self, *args ):
        if not is_empty_val(self.an_folder.get()):
            if self.an_folder_frame.is_blinking():
                self.on_entry_an_folder()

    def an_folder_check( self ):
        # TODO: include these checks in a custom widget class
        ok_ = False
        an_folder = self.an_folder.get()
        if is_empty_val( an_folder ):
            self.set_an_folder_error()
            MessageBox.showerror( "Analysis Folder Error", "The Analysis Result Folder is required.", parent=self )
        else:
            from GuiParts.NonAsciiPaths import nonascii_path_check
            if not nonascii_path_check(an_folder, self):
                return ok_

            self.an_folder_frame.stop()
            if not os.path.exists( an_folder ):
                self.set_an_folder_error()
                yn = MessageBox.askyesno( "Analysis Folder Error",
                            "'%s' does not exist.\nWould you like to make the folder?" % an_folder, icon="warning", parent=self )
                if yn:
                    try:
                        mkdirs_with_retry( an_folder )
                    except:
                        pass
                    if os.path.exists( an_folder ):
                        ok_ = True
            elif not os.path.isdir( an_folder ):
                self.set_an_folder_error()
                MessageBox.showerror( "Analysis Folder Error", "'%s' is not a folder." % an_folder, parent=self )
            elif self.check_if_like_report_folder( an_folder ):
                self.set_an_folder_error()
                yn = MessageBox.askyesno( "Analysis Folder Error",
                                            ( "'%s' seems like a used sub-folder.\n"
                                             + "Are you sure this is an appropriate selection?" )
                                            % an_folder, parent=self )
                if yn:
                    ok_ = True
            else:
                ok_ = True

        if ok_:
            self.an_folder_entry.config( fg='black' )
        else:
            self.an_folder_entry.focus_force()

        return ok_

    def check_if_like_report_folder( self, an_folder ):
        logfile = an_folder + '/molass.log'
        repfile = an_folder + '/analysis_report.xlsx'
        return os.path.exists( logfile ) and  os.path.exists( repfile )

    def set_an_folder_error( self ):
        self.an_folder_entry.config( fg='red' )

    def new_analysis( self ):
        reuse = get_setting("reuse_analysis_folder")
        if reuse:
            analysis_name = get_setting('analysis_name')
            analysis_folder = get_setting('analysis_folder')
            if analysis_folder is None:
                pass
            else:
                return analysis_name, analysis_folder

        try:
            analysis_name, analysis_folder = self.make_analysis_folder()
            self.analysis_name.set( analysis_name )
            set_setting( 'analysis_name', analysis_name )
            set_setting( 'analysis_folder', analysis_folder )
            return analysis_name, analysis_folder
        except:
            print( sys.exc_info() )
            return

    def get_analysis_name(self):
        # this method is for use in testing
        an_folder = self.an_folder.get()
        analysis_name = self.analysis_name.get()
        return self.auto_number_analysis_name(an_folder, analysis_name)

    def make_analysis_folder( self, format_check_only=False ):
        an_folder = self.an_folder.get()
        analysis_name = self.analysis_name.get()
        self.auto_number_analysis_name( an_folder, analysis_name ) # this is just for check if it is propperly named

        def get_analysis_folder( out_folder, analysis_name ):
            if is_empty_val( out_folder ):
                return None
            else:
                analysis_folder = os.path.join( out_folder, analysis_name )
                return analysis_folder.replace( '\\', '/' )

        analysis_folder = get_analysis_folder( an_folder, analysis_name )

        if not format_check_only:

            if self.auto_number.get() == 1:
                if not is_empty_val( analysis_folder ):
                    while os.path.exists( analysis_folder ) and not is_almost_empty_dir( analysis_folder ):
                        analysis_name = self.auto_number_analysis_name( an_folder, analysis_name )
                        analysis_folder = get_analysis_folder( an_folder, analysis_name )

            # print( 'make_analysis_folder: analysis_folder=', analysis_folder, type(analysis_folder) )
            if not is_empty_val( analysis_folder ) and not os.path.exists( analysis_folder ):
                mkdirs_with_retry( analysis_folder )

        return analysis_name, analysis_folder

    def auto_number_analysis_name( self, an_folder, analysis_name ):
        m = analysis_name_re.match( analysis_name )
        if m:
            first   = m.group( 1 )
            number  = m.group( 2 )
            third   = m.group( 3 )
            len_ = len( number )
            name_ = first + '%0*d' % ( len_, int( number ) ) + third
            if not is_empty_val( an_folder ):
                folder = an_folder + '/' + name_
                if os.path.exists( folder ) and not is_almost_empty_dir( folder ):
                    name_ = first + '%0*d' % ( len_, int( number ) + 1 ) + third
            self.analysis_name_entry.config( fg='black' )
            # print( 'new name=', name_ )
            return name_
        else:
            self.analysis_name_entry.config( fg='red' )
            MessageBox.showerror(
                'Naming Error',
                "Analysis Name '%s' is not expected. Name it with a numeric postfix such as 'analysis-000'." % analysis_name,
                parent=self,
                )
            assert( False )

    def analysis_name_entry_focus_out( self, event ):
        # print( 'analysis_name_entry_focus_out' )
        self.update_plot_button_state()

    def update_analysis_folder(self):
        analysis_folder = get_setting('analysis_folder')
        _, analysis_folder_ = self.make_analysis_folder(format_check_only=True)
        if analysis_folder != analysis_folder_:
            self.new_analysis()

    def prepare_serial_data( self ):
        self.tmp_logger.info('preparing serial data.')
        self.prepare_start_time = time.time()
        self.serial_data = self.loader.get_current_object()
        # note that data are not yet deepcopied at this stage
        self.abnomality_dialog = None
        if not self.use_mtd_conc:
            exclude_abnormality( self.serial_data, self.file_info_table, self.tmp_logger, dialog=self )

        if self.serial_data.has_excluded_xray_elutions:
            self.loader.memorize_exclusion( self.serial_data )

        if self.serial_data.is_serial():
            self.pre_recog = PreliminaryRecognition(self.serial_data)
        else:
            self.pre_recog = None
            return

        try:
            self.fig_frame.set_data(self.pre_recog)
        except Exception as exc:
            if not self.use_mtd_conc:
                raise exc

        eno = self.serial_data.xray_curve.primary_peak_i

        try:
            self.file_info_table.select_row(eno)
        except:
            log_exception(self.tmp_logger, "prepare_serial_data: self.file_info_table.select_row(eno)")

        self.draw_figure(eno)
        self.cfs_entry.update_path_length()

        self.dataset_is_ready = True
        preparing_time = time.time() - self.prepare_start_time
        self.tmp_logger.info('data set is ready. it took %.3g seconds for preparing.', preparing_time)
        self.update_button_states()

        self.measured_data = MeasuredData(None, sd=self.serial_data, pre_recog=self.pre_recog)
        updated = self.measured_data.update_picking_params(self.tmp_logger)
        # TODO: update self.fig_frame
        if updated:
            self.fig_frame.update_elution_curve()
        self.tmp_logger.info('created measured_data.')

        if not self.lacking_q_values_warned:
            if get_setting('found_lacking_q_values'):
                self.lacking_q_values_warned = True
                MessageBox.showwarning( "Lacking Q-values",
                    "There have been found lacking Q-values\n"
                    "in some of the input files.\n"
                    "Make sure to confirm them in the molass.log",
                    parent=self )

    def re_prepare_serial_data( self ):
        save_items = ["uv_device_no", "path_length", "beamline_name", "xray_baseline_type"]   # any other?
        saved_values = []
        for name in save_items:
            saved_values.append(get_setting(name))      # save

        self.clear_entries()

        for name, value in zip(save_items, saved_values):
            set_setting(name, value)                    # restore

        if self.dataset_is_ready:
            self.dataset_is_ready = False
            self.update_all_button_states()
            self.tmp_logger.info("re-preparing serial data after user's settings change.")
            self.after(500, self.prepare_serial_data)

    def is_reloadable(self):
        return self.dataset_is_ready

    def draw_figure( self, selected=None ):
        if self.serial_data is None or not self.serial_data.is_serial():
            return

        self.update()   # seems to be required to be sure to get the current self.an_folder value

        if is_empty_val( self.an_folder.get() ):
            self.set_state_guide_message("Specify output folder.")
            self.an_folder_frame.start()
            # self.an_folder_entry.config(bg='yellow')
            return

        self.fig_frame.draw_figure(selected)
        self.set_state_guide_message('Press [▶ Analysis start] or click to select.', fg='green')
        self.an_folder_frame.stop()
        self.update()   # seems to be required sometimes (not always)

    def wait_until_the_data_is_ready( self ):
        while self.loader.is_busy:
            self.update()
            print('waiting for the completion of data loading')
            time.sleep(1)

    def detected_abnomality( self ):
        if self.abnomality_dialog is None:
            ret = False
        else:
            self.tmp_logger.warning( 'detected abnomality(bubles?) in the input data.' )
            ret = True
        return ret

    def get_range_type( self ):
        return get_setting( 'range_type' )

    def analyze_all( self ):
        from molass_legacy.SerialAnalyzer.AbnormalityCheck import update_abnormality_fix_state

        if not self.check_all_setting_info():
            self.analysis_button.grid( row=0, column=0, padx=5 )
            return

        try:
            analysis_name, analysis_folder = self.make_analysis_folder()
            if os.path.exists( analysis_folder ) and not is_almost_empty_dir( analysis_folder ):
                self.new_analysis()
        except:
            print( sys.exc_info() )
            return

        fully_automatic = self.fully_automatic.get()
        if fully_automatic:
            ret = MessageBox.askyesno("Fully Automatic Run Confirmation",
                    '"Fully automatic" execution may take several minutes.\n'
                    + 'Are you sure to proceed?',
                    parent=self)
            if not ret:
                return

        # TODO: if using the same in_folder, ask if the previous params should be retained or discarded

        self.analysis_name.set( analysis_name )

        set_setting( 'analysis_folder', analysis_folder )
        set_setting( 'analysis_name', analysis_name )
        set_setting( 'result_book', self.result_book.get() )
        set_setting( 'fully_automatic', fully_automatic )
        if self.enable_synthesized_lrf:
            set_setting( 'synthesized_lrf', self.synthesized_lrf.get() )

        self.update_setting_info()
        self.update_plot_button_state()

        self.analyzer.change_log_to(analysis_folder)

        update_abnormality_fix_state(self.pre_recog.get_pre_recog_copy(), self.file_info_table, self.analyzer.app_logger, dialog=self)

        def v1_exec_closure():
            self.set_state_guide_message("Preparing for Mapping Dialog.")
            self.analyzer.do_analysis(self.serial_data, self.pre_recog, self.measured_data, analysis_folder, analysis_name)
            self.set_state_guide_message("V1 manipulation done.")

        self.exec_wrapper(v1_exec_closure, "do_analysis")

    def exec_wrapper(self, exec_closure, exec_name):
        # this wrapper is intended to unify the button state control

        self.is_executing = True
        self.update_button_states()
        self.update_analysis_button_state()
        self.update()

        exec_closure()

        self.reset_tmp_logger(exec_name)
        self.is_executing = False
        try:
            self.update_button_states()
            self.update_analysis_button_state()
        except:
            """
            _tkinter.TclError: invalid command name ".!guimain.!frame3.!frame.!frame.!frame.!frame.!checkbutton"
            probabply harmless
            """
            pass

    def reset_tmp_logger(self, process_name):
        self.tmp_logger = Logger()
        self.tmp_logger.info("tmp_logger has been created after %s in " % process_name + get_version_string(cpuid=True))

    def check_all_setting_info( self ):
        # for method in [ self.in_folder_entry.check, self.uv_folder_entry.check, self.an_folder_check, self.check_atsas_executability ]:
        for method in [ self.in_folder_entry.check, self.uv_folder_entry.check, self.an_folder_check ]:
            ok_ = method()
            if not ok_:
                return False

        return True

    def check_atsas_executability(self):
        from molass_legacy.ATSAS.ExecCheck import atsas_exec_check
        return atsas_exec_check(self)

    def update_setting_info( self, clear_temp=False ):
        set_setting( 'in_folder', self.in_folder.get() )
        set_setting( 'uv_folder', self.uv_folder.get() )
        set_setting( 'uv_file',   self.uv_file.get() )
        set_setting( 'an_folder', self.an_folder.get() )
        set_setting( 'auto_number', self.auto_number.get() )
        self.cfs_entry.apply_entries()
        self.save_settings(clear_temp=clear_temp)

    def update_uv_folder_entry_state( self ):
        self.use_xray_conc = get_setting( 'use_xray_conc' ) == 1
        self.use_mtd_conc = get_setting( 'use_mtd_conc' ) == 1

        uv_folder_entry = self.uv_folder_entry.entry
        if self.use_xray_conc or self.use_mtd_conc:
            in_folder = self.in_folder.get()
            if not is_empty_val( in_folder ):
                set_setting( 'file_extension', self.file_extension.get() )
                # TODO: make this symmetric with load_from_folders
                self.loader.load_xray_data_only( in_folder )
                self.loader.wait_until_ready()
            if self.use_xray_conc:
                if self.loader.has_enough_num_files():
                    text = 'No UV data; using Xray-proportional data instead'
                else:
                    # self.fig_frame.clear_figures("Guinier analysis only is\navaiblable for non-serial data.")
                    text = 'No UV data; using non-serial data'
                uv_folder_entry.config( state=Tk.NORMAL )
                uv_folder_entry.insert(0, text)
                uv_folder_entry.config( state='readonly' )
                uv_folder_entry.config( fg='orange' )
                uv_folder_entry.config( justify='center' )
                self.uv_file_entry.config( state=Tk.DISABLED )
                normal_state = False
            else:
                normal_state = True
        else:
            normal_state = True

        if normal_state:
            uv_folder_entry.config( state=Tk.NORMAL )
            uv_folder_entry.config( fg='black' )
            uv_folder_entry.config( justify='left' )
            self.uv_file_entry.config( state=Tk.NORMAL )

        self.update_analysis_button_state()

    def doing_mfc(self):
        return self.use_mtd_conc == 1

    def plot_results( self ):
        print( 'plot_results' )
        from molass_legacy.GuinierAnalyzer.SimpleGuinierAnalyzer import SimpleGuinierAnalyzer
        from molass_legacy.GuinierAnalyzer.SimpleGuinierFolderAnalyzer import SimpleGuinierFolderAnalyzer
        in_folder = self.in_folder.get()
        sg_analyzer = SimpleGuinierAnalyzer( self.maintenance_log )
        report_folder = '/'.join( self.maintenance_log.split('/')[:-2]  )
        fo_analyzer = SimpleGuinierFolderAnalyzer( self, -1, in_folder,
                        sg_analyzer.folder_infos[0],
                        sg_analyzer.x, sg_analyzer.z, sg_analyzer.r,
                        sg_analyzer.type_names,
                        report_folder=report_folder )

        fo_analyzer.show()

    def update_plot_button_state( self ):
        maintenance_mode = get_setting( 'maintenance_mode' )
        if maintenance_mode == 0:
            self.plot_button.grid_forget()
            self.maintenance_log = None
        else:
            self.plot_button.grid( row=0, column=1, padx=5 )
            self.maintenance_log = self.make_maintenance_log_path()
        if self.same_folder_in_maintenance_log():
            state = Tk.NORMAL
        else:
            state = Tk.DISABLED
        self.plot_button.config( state=state )

    def same_folder_in_maintenance_log( self ):
        if self.maintenance_log is None:
            return False
        if not os.path.exists( self.maintenance_log ):
            return False

        fh = open( self.maintenance_log )
        line = fh.readline()
        fh.close()

        folder = line.replace( 'Doing ', '' )
        return folder[0:-1] == self.in_folder.get()

    def make_maintenance_log_path( self ):
        an_folder       = self.an_folder.get()
        analysis_name   = self.analysis_name.get()
        if is_empty_val(an_folder) or is_empty_val(analysis_name):
            return None

        return an_folder + '/' + analysis_name + '/.temp/maintenance.log'

    def save_settings( self, clear_temp=False ):
        restore_default_setting( 'file_extension' )
        if clear_temp:
            clear_temporary_settings()
        try:
            save_settings()
            AutorgKek.Settings.save_settings()    # for saving AutorgKek settings
        except:
            pass

    def set_dev_setting( self, item, value ):
        set_dev_setting( item, value )

    def set_setting( self, item, value ):
        set_setting( item, value )

    def _exit(self):
        if self.cleaner is not None:
            if self.tmp_logger is None:
                # this can be avoided; temp fix
                pass
            else:
                self.tmp_logger.info("os._exit(0)")
            os._exit(0)

    def quit( self, immediately=False ):
        # print( 'quit' )

        if immediately:
            self.parent.quit()
            self.destroy()
            self._exit()
            return

        reply = MessageBox.askokcancel(
                "Quit Confirmation",
                "Do you really want to quit?",
                parent=self,
                )
        if not reply:
            return

        from molass_legacy.DENSS.DenssManagerDialog import terminate_manager
        ret = terminate_manager(self)
        if not ret:
            return

        from molass_legacy._MOLASS.Processes import terminate_all_processes
        terminate_all_processes()

        # print( self.geometry() )
        # set_setting( 'main_geometry',   self.geometry()          )

        """
        if self.settings_have_been_changed:
            reply = MessageBox.askyesno(
                    "Setting Info Save Question",
                    "Save the setting information?",
                    parent=self,
                    )
            if reply:
                save_settings()
        """
        save_settings(clear_no_save_items=True)
        self.update_setting_info( clear_temp=True )
        self.parent.quit()
        self.destroy()
        self._exit()

    def is_placed_foremost( self ):
        # TODO
        # num_stacked = len( self.tk.eval( 'wm stackorder '+str(self.top_widget) ).split(' ') )
        # return num_stacked == 1
        return True

    def _on_control_c( self, event ):
        w = self.focus_get()
        # print( 'table=', self.file_info_table )
        # print( 'w=', w )
        if str(w).find( str(self.file_info_table) ) == 0:
            self.file_info_table.table._on_control_c( event )
        else:
            # TODO:  w.selection
            pass

    def _on_click( self, event ):
        # the follwing order is essential to keep exluded marks (i.e., 'X')
        # unchanged in case of non-cell clicks.
        self.file_info_table.table._on_click( event )
        self.file_info_table._on_click( event )

    def _on_enter( self, event ):
        # TODO: only when changed
        # self.on_entry_in_folder()
        if not self.is_placed_foremost(): return

        print( 'ENTER in main window' )
        state = self.analysis_button.cget( 'state' )
        if state == Tk.NORMAL:
            self.analyze_all()
        else:
            self.in_folder_entry.on_entry()

    def _on_escape( self, event ):
        if not self.is_placed_foremost(): return

        self.quit()

    def do_simple_guinier( self, i ):
        from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
        from molass_legacy.GuinierAnalyzer.AutorgKekAdapter import AutorgKekAdapter
        print( 'do_simple_guinier: i=', i )

        data, file, _, n = self.file_info_table.get_data_array( i )

        guinier = SimpleGuinier( data )
        adapter = AutorgKekAdapter( None, guinier=guinier )
        adapter_result = adapter.run()
        return adapter_result

    def show_threedim_dialog(self, debug=False):
        if debug:
            from importlib import reload
            import Tools.ThreeDimViewer
            reload(Tools.ThreeDimViewer)
        from molass_legacy.Tools.ThreeDimViewer import ThreeDimViewer
        self.threedim_dialog = ThreeDimViewer(self.parent, self.measured_data)
        self.threedim_dialog.show()

    def show_lpm_inspector(self):
        from molass_legacy.Baseline.LpmInspect import LpmInspector
        self.baseline_inspector = LpmInspector(self.parent, self.measured_data)
        self.baseline_inspector.show()

    def disable_uv_data_tracer(self, *args):
        if not self.disable_uv_data_tracing:
            print('skipping disable_uv_data_tracer')
            return

        self.disable_uv_data_tracer_busy = True

        disable_uv_data = self.disable_uv_data.get()
        set_setting('disable_uv_data', disable_uv_data)
        if disable_uv_data:
            self.uv_folder.set('')
            self.uv_file.set('')
        self.uv_file_info_changed = True
        self.refresh(keeping_setting_items=['disable_uv_data'])

        self.disable_uv_data_tracer_busy = False

    def get_setting_for_test(self, item):
        return get_setting(item)



"""
    ModelParams.EghPlotUtils.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Optimizer.TheDebugUtils import convert_score_list, plot_scores
from molass_legacy.GuinierTools.RgCurveUtils import plot_rg_curves
from molass_legacy.Optimizer.FvScoreConverter import convert_score

def plot_objective_state(score_list_pair, fv,
        lrf_info,
        overlap, rg_curve, rg_params,
        score_names,
        fig_info, axis_info,
        func, params,
        ):
    if func is None or params is None:
        pass
    else:
        split_params = func.split_params_simple(params)
        xr_params, xr_baseparams, rg_params_, (a, b), uv_params, uv_baseparams, (c, d), seccol_params = split_params[0:8]   # i.e., without (Ti, Np)
        mapped_uv_params = []
        for (h, m, s, t), uv_h in zip(xr_params, uv_params):
            mapped_uv_params.append((uv_h, a*m + b, a*s, a*t))

    uv_x = lrf_info.uv_x
    uv_y = lrf_info.uv_y
    uv_cy_list = lrf_info.get_uv_cy_list()
    uv_ty = lrf_info.uv_ty
    x = lrf_info.x
    y = lrf_info.y
    xr_cy_list = lrf_info.get_xr_cy_list()
    xr_ty = lrf_info.xr_ty

    num_components = lrf_info.get_num_substantial_components()
    score_list = convert_score_list(score_list_pair)

    if fig_info is None:
        in_folder, seeds, result = [None]*3
    else:
        in_folder, seeds, result = fig_info

    if axis_info is None:
        plt.push()
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7))
        axt = ax2.twinx()
        axt.grid(False)
    else:
        if len(axis_info) == 2:
            fig, (ax1, ax2, ax3, axt) = axis_info
        else:
            assert False

    if result is not None:
        func_name =  "" if func is None else " " + func.get_name()
        fig.suptitle("Objective Function%s Debug Plot for %s, SV=%.3g" % (func_name, in_folder, convert_score(result.fun)), fontsize=20)

    if ax1 is not None:
        ax1.plot(uv_x, uv_y, color='blue')
        points = lrf_info.estimate_uv_peak_points()
        for k, uv_cy in enumerate(uv_cy_list):
            if k < num_components:
                style = ':'
                label = 'component-%d' % (k+1)
                color = None
            else:
                style = '-'
                label = 'baseline'
                color = 'red'
            ax1.plot(uv_x, uv_cy, style, color=color, label=label)
        ax1.plot(uv_x, uv_ty, ':', color='red', lw=2, label='total')
        ax1.plot(*points.T, "o", color="yellow")
        ax1.legend()

    if ax2 is not None:
        ax2.plot(x, y, color='orange')
        num_drawn_flags = 0
        points = lrf_info.estimate_xr_peak_points()
        for k, xr_cy in enumerate(xr_cy_list):
            if k < num_components:
                style = ':'
                label = 'component-%d' % (k+1)
                color = None
            elif k == num_components:
                style = '-'
                label = 'baseline'
                color = 'red'
            else:
                style = ':'
                label = 'effect of ii' if num_drawn_flags == 0 else None
                color = 'pink'
                num_drawn_flags += 1
            ax2.plot(x, xr_cy, style, color=color, label=label)
        ax2.plot(x, xr_ty, ':', color='red', lw=2, label='total')
        ax2.plot(*points.T, "o", color="yellow")

        if False:
            # consider overlaps later
            y1 = np.zeros(len(x))
            ax2.fill_between(x, y1, overlap, fc='pink', alpha=0.2)

        ax2.legend(loc='upper right')

    if axt is not None:
        assert ax2 is not None
        rg_params_ = lrf_info.get_valid_rgs(rg_params)
        plot_rg_curves(axt, points[:,1], rg_params_, x, xr_cy_list, xr_ty, rg_curve)

    if ax3 is not None:
        plot_scores(ax3, score_list, score_names)

    if axis_info is None:
        fig.tight_layout()
        if result is not None:
            fig.subplots_adjust(top=0.9)
        debug_fv = plt.show()
        plt.pop()
    else:
        debug_fv = None

    return debug_fv
